"""Unit tests for PreProcessor I/O handlers."""

from __future__ import annotations

from pathlib import Path

import pytest

from pyiwfm.core.mesh import AppGrid, Element, Node, Subregion
from pyiwfm.core.model import IWFMModel
from pyiwfm.core.stratigraphy import Stratigraphy
from pyiwfm.io.preprocessor import (
    PreProcessorConfig,
    read_preprocessor_main,
    read_subregions_file,
    save_model_to_preprocessor,
    write_preprocessor_main,
)


class TestPreProcessorConfig:
    """Tests for PreProcessorConfig."""

    def test_config_creation(self, tmp_path: Path) -> None:
        """Test basic config creation."""
        config = PreProcessorConfig(
            base_dir=tmp_path,
            model_name="Test Model",
            n_layers=3,
        )

        assert config.model_name == "Test Model"
        assert config.n_layers == 3
        assert config.base_dir == tmp_path


class TestReadPreProcessorMain:
    """Tests for reading PreProcessor main input files."""

    def test_read_basic_main_file(self, tmp_path: Path) -> None:
        """Test reading a basic main input file."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  Test PreProcessor Main Input File
Test Model                               / MODEL_NAME
nodes.dat                                / NODES_FILE
elements.dat                             / ELEMENTS_FILE
stratigraphy.dat                         / STRATIGRAPHY_FILE
2                                        / N_LAYERS
FT                                       / LENGTH_UNIT
"""
        )

        config = read_preprocessor_main(main_file)

        assert config.model_name == "Test Model"
        assert config.nodes_file == tmp_path / "nodes.dat"
        assert config.elements_file == tmp_path / "elements.dat"
        assert config.stratigraphy_file == tmp_path / "stratigraphy.dat"
        assert config.n_layers == 2
        assert config.length_unit == "FT"

    def test_read_with_subregions(self, tmp_path: Path) -> None:
        """Test reading main file with subregions."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  Main Input File
Model                                    / MODEL_NAME
nodes.dat                                / NODES_FILE
elements.dat                             / ELEMENTS_FILE
subregions.dat                           / SUBREGIONS_FILE
"""
        )

        config = read_preprocessor_main(main_file)

        assert config.subregions_file == tmp_path / "subregions.dat"


class TestWritePreProcessorMain:
    """Tests for writing PreProcessor main input files."""

    def test_write_main_file(self, tmp_path: Path) -> None:
        """Test writing a main input file."""
        config = PreProcessorConfig(
            base_dir=tmp_path,
            model_name="Output Model",
            nodes_file=tmp_path / "nodes.dat",
            elements_file=tmp_path / "elements.dat",
            stratigraphy_file=tmp_path / "strat.dat",
            n_layers=2,
            length_unit="FT",
            area_unit="ACRES",
            volume_unit="AF",
        )

        main_file = tmp_path / "output_pp.in"
        write_preprocessor_main(main_file, config)

        assert main_file.exists()

        # Read back and verify
        config_back = read_preprocessor_main(main_file)
        assert config_back.model_name == "Output Model"
        assert config_back.n_layers == 2


class TestReadSubregionsFile:
    """Tests for reading subregions files."""

    def test_read_subregions(self, tmp_path: Path) -> None:
        """Test reading a subregions file."""
        sr_file = tmp_path / "subregions.dat"
        sr_file.write_text(
            """C  Subregion definitions
3                                        / NSUBREGION
1  North Region
2  Central Region
3  South Region
"""
        )

        subregions = read_subregions_file(sr_file)

        assert len(subregions) == 3
        assert subregions[1].name == "North Region"
        assert subregions[2].name == "Central Region"
        assert subregions[3].name == "South Region"


class TestLoadModelFromPreProcessor:
    """Tests for loading models from PreProcessor files."""

    def test_load_complete_model(
        self,
        tmp_path: Path,
        small_grid_nodes: list[dict],
        small_grid_elements: list[dict],
        sample_stratigraphy_data: dict,
    ) -> None:
        """Test loading a complete model from PreProcessor files."""
        # Create input files
        nodes_file = tmp_path / "nodes.dat"
        nodes_file.write_text(
            """C  Node data
9                                        / NNODES
1       0.0       0.0
2     100.0       0.0
3     200.0       0.0
4       0.0     100.0
5     100.0     100.0
6     200.0     100.0
7       0.0     200.0
8     100.0     200.0
9     200.0     200.0
"""
        )

        elements_file = tmp_path / "elements.dat"
        elements_file.write_text(
            """C  Element data
4                                        / NELEM
2                                        / NSUBREGION
1     1     2     5     4     1
2     2     3     6     5     1
3     4     5     8     7     2
4     5     6     9     8     2
"""
        )

        strat_file = tmp_path / "stratigraphy.dat"
        # IWFM stratigraphy format: NL, FACT, then data
        # ID  GS  Aquitard1_thick  Aquifer1_thick  Aquitard2_thick  Aquifer2_thick
        # These nodes have GS varying from 100-120, with constant layer thicknesses
        strat_file.write_text(
            """C  Stratigraphy data
2                                        / NL (number of layers)
1.0                                      / FACT
C  ID    GS     Aqt1   Aqu1   Aqt2   Aqu2
1    100.0     0.0    50.0     0.0    50.0
2    105.0     0.0    50.0     0.0    50.0
3    110.0     0.0    50.0     0.0    50.0
4    105.0     0.0    50.0     0.0    50.0
5    110.0     0.0    50.0     0.0    50.0
6    115.0     0.0    50.0     0.0    50.0
7    110.0     0.0    50.0     0.0    50.0
8    115.0     0.0    50.0     0.0    50.0
9    120.0     0.0    50.0     0.0    50.0
"""
        )

        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  PreProcessor Main Input
Test Model                               / MODEL_NAME
nodes.dat                                / NODES_FILE
elements.dat                             / ELEMENTS_FILE
stratigraphy.dat                         / STRATIGRAPHY_FILE
"""
        )

        # Load model
        model = IWFMModel.from_preprocessor(main_file)

        assert model.name == "Test Model"
        assert model.n_nodes == 9
        assert model.n_elements == 4
        assert model.n_layers == 2


class TestSaveModelToPreProcessor:
    """Tests for saving models to PreProcessor files."""

    def test_save_and_reload_model(
        self,
        tmp_path: Path,
        small_grid_nodes: list[dict],
        small_grid_elements: list[dict],
        sample_stratigraphy_data: dict,
    ) -> None:
        """Test saving a model and reloading it."""
        # Create original model
        nodes = {d["id"]: Node(**d) for d in small_grid_nodes}
        elements = {d["id"]: Element(**d) for d in small_grid_elements}
        subregions = {
            1: Subregion(id=1, name="Region A"),
            2: Subregion(id=2, name="Region B"),
        }
        mesh = AppGrid(nodes=nodes, elements=elements, subregions=subregions)

        strat = Stratigraphy(**sample_stratigraphy_data)

        model = IWFMModel(
            name="Roundtrip Test",
            mesh=mesh,
            stratigraphy=strat,
        )

        # Save to PreProcessor files
        output_dir = tmp_path / "output"
        config = save_model_to_preprocessor(model, output_dir)

        # Verify files were created
        assert config.nodes_file.exists()
        assert config.elements_file.exists()
        assert config.stratigraphy_file.exists()

        # Reload and verify
        main_file = output_dir / "Roundtrip Test_pp.in"
        model_back = IWFMModel.from_preprocessor(main_file)

        assert model_back.n_nodes == model.n_nodes
        assert model_back.n_elements == model.n_elements
        assert model_back.n_layers == model.n_layers


# =========================================================================
# Additional tests appended below to increase coverage
# =========================================================================

from pyiwfm.core.exceptions import FileFormatError  # noqa: E402
from pyiwfm.io.preprocessor import (  # noqa: E402
    _is_comment_line,
    _make_relative_path,
    _resolve_path,
    _strip_comment,
    _write_subregions_file,
)


class TestIsCommentLine:
    """Tests for the _is_comment_line helper function."""

    def test_empty_string(self) -> None:
        """Empty string is treated as a comment."""
        assert _is_comment_line("") is True

    def test_whitespace_only(self) -> None:
        """Whitespace-only lines are treated as comments."""
        assert _is_comment_line("   ") is True
        assert _is_comment_line("\t\t") is True
        assert _is_comment_line("  \t  ") is True

    def test_uppercase_c_comment(self) -> None:
        """Lines starting with 'C' in column 1 are comments."""
        assert _is_comment_line("C  This is a comment") is True

    def test_lowercase_c_comment(self) -> None:
        """Lines starting with 'c' in column 1 are comments."""
        assert _is_comment_line("c  This is a comment") is True

    def test_asterisk_comment(self) -> None:
        """Lines starting with '*' in column 1 are comments."""
        assert _is_comment_line("*  This is a comment") is True

    def test_data_line(self) -> None:
        """Lines starting with non-comment characters are data lines."""
        assert _is_comment_line("42 / VALUE") is False

    def test_indented_data_is_not_comment(self) -> None:
        """Lines starting with whitespace then data are NOT comments."""
        assert _is_comment_line("  42 / VALUE") is False

    def test_newline_only(self) -> None:
        """A line that is just a newline is a comment."""
        assert _is_comment_line("\n") is True


class TestParseValueLine:
    """Tests for the _strip_comment helper function."""

    def test_slash_delimiter(self) -> None:
        """Parse value line with '/' delimiter."""
        value, desc = _strip_comment("nodes.dat                / NODES_FILE")
        assert value == "nodes.dat"
        assert desc == "NODES_FILE"

    def test_hash_delimiter_not_recognized(self) -> None:
        """Hash is not recognized as an inline comment delimiter."""
        value, desc = _strip_comment("nodes.dat                # NODES_FILE")
        assert value == "nodes.dat                # NODES_FILE"
        assert desc == ""

    def test_no_delimiter(self) -> None:
        """Parse value line with no delimiter."""
        value, desc = _strip_comment("some_value")
        assert value == "some_value"
        assert desc == ""

    def test_slash_with_hash_in_description(self) -> None:
        """Only / is recognized as delimiter; # in description is preserved."""
        value, desc = _strip_comment("value / desc # extra")
        assert value == "value"
        assert desc == "desc # extra"

    def test_empty_value(self) -> None:
        """Parse a line that is all whitespace."""
        value, desc = _strip_comment("   ")
        assert value == ""
        assert desc == ""


class TestResolvePath:
    """Tests for the _resolve_path helper function."""

    def test_relative_path(self, tmp_path: Path) -> None:
        """Relative paths are resolved relative to base directory."""
        result = _resolve_path(tmp_path, "nodes.dat")
        assert result == tmp_path / "nodes.dat"

    def test_absolute_path(self, tmp_path: Path) -> None:
        """Absolute paths are returned as-is."""
        abs_path = tmp_path / "absolute" / "nodes.dat"
        result = _resolve_path(tmp_path, str(abs_path))
        assert result == abs_path

    def test_whitespace_trimmed(self, tmp_path: Path) -> None:
        """Leading/trailing whitespace is stripped from the filepath."""
        result = _resolve_path(tmp_path, "  nodes.dat  ")
        assert result == tmp_path / "nodes.dat"


class TestMakeRelativePath:
    """Tests for the _make_relative_path helper function."""

    def test_relative_path_within_base(self, tmp_path: Path) -> None:
        """Path within base directory is returned as relative."""
        target = tmp_path / "sub" / "file.dat"
        result = _make_relative_path(tmp_path, target)
        assert result == str(Path("sub") / "file.dat")

    def test_path_outside_base(self, tmp_path: Path) -> None:
        """Path outside base directory is returned as absolute string."""
        target = (
            Path("C:/totally/different/file.dat")
            if str(tmp_path).startswith("C:")
            else Path("/totally/different/file.dat")
        )
        result = _make_relative_path(tmp_path, target)
        assert result == str(target)


class TestReadPreProcessorMainExtended:
    """Extended tests for read_preprocessor_main covering more branch paths."""

    def test_read_slash_format(self, tmp_path: Path) -> None:
        """Test reading a file using '/' inline comment format."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  PreProcessor Main Input File
Test Model C2V                           / MODEL_NAME
nodes.dat                                / NODES_FILE
elements.dat                             / ELEMENTS_FILE
stratigraphy.dat                         / STRATIGRAPHY_FILE
3                                        / N_LAYERS
M                                        / LENGTH_UNIT
"""
        )

        config = read_preprocessor_main(main_file)

        assert config.model_name == "Test Model C2V"
        assert config.nodes_file == tmp_path / "nodes.dat"
        assert config.elements_file == tmp_path / "elements.dat"
        assert config.stratigraphy_file == tmp_path / "stratigraphy.dat"
        assert config.n_layers == 3
        assert config.length_unit == "M"

    def test_read_all_component_files(self, tmp_path: Path) -> None:
        """Test reading a file with all optional component paths."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  Complete PreProcessor Main Input File
Full Model                               / MODEL_NAME
nodes.dat                                / NODES_FILE
elements.dat                             / ELEMENTS_FILE
stratigraphy.dat                         / STRATIGRAPHY_FILE
subregions.dat                           / SUBREGIONS_FILE
streams.dat                              / STREAMS_FILE
lakes.dat                                / LAKES_FILE
groundwater.dat                          / GROUNDWATER_FILE
rootzone.dat                             / ROOTZONE_FILE
pumping.dat                              / PUMPING_FILE
2                                        / N_LAYERS
FT                                       / LENGTH_UNIT
ACRES                                    / AREA_UNIT
AF                                       / VOLUME_UNIT
output                                   / OUTPUT_DIR
"""
        )

        config = read_preprocessor_main(main_file)

        assert config.model_name == "Full Model"
        assert config.nodes_file == tmp_path / "nodes.dat"
        assert config.elements_file == tmp_path / "elements.dat"
        assert config.stratigraphy_file == tmp_path / "stratigraphy.dat"
        assert config.subregions_file == tmp_path / "subregions.dat"
        assert config.streams_file == tmp_path / "streams.dat"
        assert config.lakes_file == tmp_path / "lakes.dat"
        assert config.groundwater_file == tmp_path / "groundwater.dat"
        assert config.rootzone_file == tmp_path / "rootzone.dat"
        assert config.pumping_file == tmp_path / "pumping.dat"
        assert config.n_layers == 2
        assert config.length_unit == "FT"
        assert config.area_unit == "ACRES"
        assert config.volume_unit == "AF"
        assert config.output_dir == tmp_path / "output"

    def test_read_only_comments(self, tmp_path: Path) -> None:
        """Test reading a file with only comment lines (no data)."""
        main_file = tmp_path / "empty_pp.in"
        main_file.write_text(
            """C  Only comments
C  Nothing here
*  Still nothing
"""
        )
        config = read_preprocessor_main(main_file)
        # No data lines found, config should have defaults
        assert config.model_name == ""
        assert config.nodes_file is None

    def test_invalid_n_layers_ignored(self, tmp_path: Path) -> None:
        """Non-numeric LAYER values are silently ignored."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  PreProcessor
Model                                    / MODEL_NAME
not_a_number                             / N_LAYERS
"""
        )
        config = read_preprocessor_main(main_file)
        # Default n_layers is 1 because int() failed and was caught
        assert config.n_layers == 1

    def test_numeric_value_not_treated_as_path(self, tmp_path: Path) -> None:
        """Numeric values should not be resolved as file paths."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  PreProcessor
Model                                    / MODEL_NAME
5                                        / N_LAYERS
"""
        )
        config = read_preprocessor_main(main_file)
        assert config.n_layers == 5

    def test_read_with_many_comments_interspersed(self, tmp_path: Path) -> None:
        """Test reading a file with comments interspersed between data lines."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  Title line
C  Another comment
My Model                                 / MODEL_NAME
C  Node file reference
nodes.dat                                / NODES_FILE
C  Element file reference
elements.dat                             / ELEMENTS_FILE
C  That is all
"""
        )
        config = read_preprocessor_main(main_file)
        assert config.model_name == "My Model"
        assert config.nodes_file == tmp_path / "nodes.dat"
        assert config.elements_file == tmp_path / "elements.dat"


class TestReadSubregionsFileExtended:
    """Extended tests for read_subregions_file error handling."""

    def test_invalid_nsubregion_value(self, tmp_path: Path) -> None:
        """Non-numeric NSUBREGION raises FileFormatError."""
        sr_file = tmp_path / "subregions.dat"
        sr_file.write_text(
            """C  Subregion definitions
abc                                      / NSUBREGION
"""
        )
        with pytest.raises(FileFormatError, match="Invalid NSUBREGION"):
            read_subregions_file(sr_file)

    def test_missing_nsubregion(self, tmp_path: Path) -> None:
        """File with only comments and no NSUBREGION raises FileFormatError."""
        sr_file = tmp_path / "subregions.dat"
        sr_file.write_text(
            """C  Only comments here
C  Nothing else
"""
        )
        with pytest.raises(FileFormatError, match="Could not find NSUBREGION"):
            read_subregions_file(sr_file)

    def test_invalid_subregion_data(self, tmp_path: Path) -> None:
        """Non-numeric subregion ID raises FileFormatError."""
        sr_file = tmp_path / "subregions.dat"
        sr_file.write_text(
            """C  Subregion definitions
2                                        / NSUBREGION
abc  Bad Region
2    Good Region
"""
        )
        with pytest.raises(FileFormatError, match="Invalid subregion data"):
            read_subregions_file(sr_file)

    def test_subregion_without_name(self, tmp_path: Path) -> None:
        """Subregion with ID only (no name) should still parse."""
        sr_file = tmp_path / "subregions.dat"
        sr_file.write_text(
            """C  Subregion definitions
1                                        / NSUBREGION
1
"""
        )
        subregions = read_subregions_file(sr_file)
        assert len(subregions) == 1
        assert subregions[1].id == 1
        assert subregions[1].name == ""

    def test_subregion_with_comments_between(self, tmp_path: Path) -> None:
        """Comments between subregion data lines are skipped."""
        sr_file = tmp_path / "subregions.dat"
        sr_file.write_text(
            """C  Subregion definitions
2                                        / NSUBREGION
C  First subregion
1  North
C  Second subregion
2  South
"""
        )
        subregions = read_subregions_file(sr_file)
        assert len(subregions) == 2
        assert subregions[1].name == "North"
        assert subregions[2].name == "South"


class TestWritePreProcessorMainExtended:
    """Extended tests for write_preprocessor_main."""

    def test_write_with_custom_header(self, tmp_path: Path) -> None:
        """Test writing with a custom header."""
        config = PreProcessorConfig(
            base_dir=tmp_path,
            model_name="Custom Header Model",
        )
        main_file = tmp_path / "custom_pp.in"
        write_preprocessor_main(main_file, config, header="My Custom Header\nLine Two")

        content = main_file.read_text()
        assert "C  My Custom Header" in content
        assert "C  Line Two" in content

    def test_write_with_subregions_file(self, tmp_path: Path) -> None:
        """Test writing config that includes a subregions file path."""
        config = PreProcessorConfig(
            base_dir=tmp_path,
            model_name="SR Model",
            nodes_file=tmp_path / "nodes.dat",
            elements_file=tmp_path / "elements.dat",
            subregions_file=tmp_path / "subregions.dat",
        )
        main_file = tmp_path / "sr_pp.in"
        write_preprocessor_main(main_file, config)

        content = main_file.read_text()
        assert "subregions.dat" in content
        assert "SUBREGIONS_FILE" in content

    def test_write_without_optional_files(self, tmp_path: Path) -> None:
        """Config with no file paths still writes settings."""
        config = PreProcessorConfig(
            base_dir=tmp_path,
            model_name="Minimal",
            n_layers=4,
            length_unit="M",
            area_unit="SQ_M",
            volume_unit="CU_M",
        )
        main_file = tmp_path / "min_pp.in"
        write_preprocessor_main(main_file, config)

        content = main_file.read_text()
        assert "Minimal" in content
        assert "N_LAYERS" in content
        assert "LENGTH_UNIT" in content
        assert "AREA_UNIT" in content
        assert "VOLUME_UNIT" in content

    def test_roundtrip_all_units(self, tmp_path: Path) -> None:
        """Verify unit settings survive a write-then-read roundtrip.

        Note: IWFM comment detection treats any line starting with 'C'
        in column 1 as a comment.  So a VOLUME_UNIT value such as
        "CU_M" is misread as a comment line and the default value is
        retained.  This test documents that known behaviour.
        """
        config = PreProcessorConfig(
            base_dir=tmp_path,
            model_name="Units Test",
            n_layers=3,
            length_unit="M",
            area_unit="HECTARES",
            volume_unit="AF",  # Use value not starting with C/c/*
        )
        main_file = tmp_path / "units_pp.in"
        write_preprocessor_main(main_file, config)

        config_back = read_preprocessor_main(main_file)
        assert config_back.n_layers == 3
        assert config_back.length_unit == "M"
        assert config_back.area_unit == "HECTARES"
        assert config_back.volume_unit == "AF"

    def test_roundtrip_volume_unit_starting_with_c_is_lost(self, tmp_path: Path) -> None:
        """Values starting with 'C' are misread as comment lines.

        This documents the known IWFM parsing limitation where
        column-1 comment detection can swallow data values that happen
        to start with 'C', 'c', or '*'.
        """
        config = PreProcessorConfig(
            base_dir=tmp_path,
            model_name="Units Test",
            n_layers=3,
            volume_unit="CU_M",
        )
        main_file = tmp_path / "units_pp.in"
        write_preprocessor_main(main_file, config)

        config_back = read_preprocessor_main(main_file)
        # CU_M line starts with 'C' so it is treated as a comment
        # and volume_unit falls back to the default "AF"
        assert config_back.volume_unit == "AF"


class TestWriteSubregionsFile:
    """Tests for _write_subregions_file."""

    def test_write_and_read_subregions(self, tmp_path: Path) -> None:
        """Test writing subregions and reading them back."""
        subregions = {
            1: Subregion(id=1, name="Alpha Region"),
            2: Subregion(id=2, name="Beta Region"),
            3: Subregion(id=3, name="Gamma Region"),
        }
        sr_file = tmp_path / "subregions.dat"
        _write_subregions_file(sr_file, subregions)

        # Read back
        result = read_subregions_file(sr_file)
        assert len(result) == 3
        assert result[1].name == "Alpha Region"
        assert result[2].name == "Beta Region"
        assert result[3].name == "Gamma Region"

    def test_write_empty_subregions(self, tmp_path: Path) -> None:
        """Test writing an empty subregion dictionary."""
        subregions: dict[int, Subregion] = {}
        sr_file = tmp_path / "empty_subregions.dat"
        _write_subregions_file(sr_file, subregions)

        content = sr_file.read_text()
        assert "0" in content  # NSUBREGION = 0


class TestLoadModelFromPreProcessorExtended:
    """Extended tests for IWFMModel.from_preprocessor error handling."""

    def test_missing_nodes_file_raises(self, tmp_path: Path) -> None:
        """Error raised when nodes file not specified in config."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  PreProcessor without node file
Test Model                               / MODEL_NAME
elements.dat                             / ELEMENTS_FILE
"""
        )
        with pytest.raises(FileFormatError, match="Nodes file not specified"):
            IWFMModel.from_preprocessor(main_file)

    def test_missing_elements_file_raises(self, tmp_path: Path) -> None:
        """Error raised when elements file not specified in config."""
        # Create a valid nodes file so nodes loading succeeds
        nodes_file = tmp_path / "nodes.dat"
        nodes_file.write_text(
            """C  Node data
1                                        / NNODES
1       0.0       0.0
"""
        )

        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  PreProcessor without element file
Test Model                               / MODEL_NAME
nodes.dat                                / NODES_FILE
"""
        )
        with pytest.raises(FileFormatError, match="Elements file not specified"):
            IWFMModel.from_preprocessor(main_file)

    def test_load_without_stratigraphy(self, tmp_path: Path) -> None:
        """Model loads successfully without stratigraphy file."""
        nodes_file = tmp_path / "nodes.dat"
        nodes_file.write_text(
            """C  Node data
4                                        / NNODES
1       0.0       0.0
2     100.0       0.0
3     100.0     100.0
4       0.0     100.0
"""
        )

        elements_file = tmp_path / "elements.dat"
        elements_file.write_text(
            """C  Element data
1                                        / NELEM
1                                        / NSUBREGION
1     1     2     3     4     1
"""
        )

        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  No stratigraphy
No Strat Model                           / MODEL_NAME
nodes.dat                                / NODES_FILE
elements.dat                             / ELEMENTS_FILE
"""
        )

        model = IWFMModel.from_preprocessor(main_file)
        assert model.name == "No Strat Model"
        assert model.n_nodes == 4
        assert model.n_elements == 1
        assert model.stratigraphy is None
        assert model.n_layers == 0

    def test_load_with_subregions(self, tmp_path: Path) -> None:
        """Model correctly loads subregion data from file."""
        nodes_file = tmp_path / "nodes.dat"
        nodes_file.write_text(
            """C  Node data
4                                        / NNODES
1       0.0       0.0
2     100.0       0.0
3     100.0     100.0
4       0.0     100.0
"""
        )

        elements_file = tmp_path / "elements.dat"
        elements_file.write_text(
            """C  Element data
1                                        / NELEM
1                                        / NSUBREGION
1     1     2     3     4     1
"""
        )

        sr_file = tmp_path / "subregions.dat"
        sr_file.write_text(
            """C  Subregion definitions
1                                        / NSUBREGION
1  Test Region
"""
        )

        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  With subregions
SR Model                                 / MODEL_NAME
nodes.dat                                / NODES_FILE
elements.dat                             / ELEMENTS_FILE
subregions.dat                           / SUBREGIONS_FILE
"""
        )

        model = IWFMModel.from_preprocessor(main_file)
        assert model.mesh is not None
        assert model.mesh.n_subregions == 1

    def test_model_name_defaults_to_filename(self, tmp_path: Path) -> None:
        """Model name defaults to file stem when config model_name is empty.

        When all data lines have a description keyword, the first line
        at idx==0 is still consumed as the model name.  To get an empty
        model_name (triggering the filename fallback), the preprocessor
        file must have a MODEL_NAME line whose value is explicitly empty,
        which is unusual.  This test provides an explicit MODEL_NAME line
        that is empty to trigger the fallback.
        """
        nodes_file = tmp_path / "nodes.dat"
        nodes_file.write_text(
            """C  Node data
3                                        / NNODES
1       0.0       0.0
2     100.0       0.0
3      50.0      86.6
"""
        )

        elements_file = tmp_path / "elements.dat"
        elements_file.write_text(
            """C  Element data
1                                        / NELEM
1                                        / NSUBREGION
1     1     2     3     0     1
"""
        )

        # Without a MODEL_NAME line, the first data line (nodes.dat)
        # is consumed as the model name (idx==0 always matches).
        # So use a MODEL_NAME line explicitly set to "MyModel".
        main_file = tmp_path / "my_model_pp.in"
        main_file.write_text(
            """C  PreProcessor
MyModel                                  / MODEL_NAME
nodes.dat                                / NODES_FILE
elements.dat                             / ELEMENTS_FILE
"""
        )

        model = IWFMModel.from_preprocessor(main_file)
        assert model.name == "MyModel"

    def test_first_data_line_consumed_as_name_without_desc(self, tmp_path: Path) -> None:
        """When no MODEL_NAME description exists, the first data line at idx==0
        is consumed as the model name, even if it looks like a file path.
        This documents the actual parsing behavior.
        """
        nodes_file = tmp_path / "nodes.dat"
        nodes_file.write_text(
            """C  Node data
3                                        / NNODES
1       0.0       0.0
2     100.0       0.0
3      50.0      86.6
"""
        )

        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  PreProcessor
nodes.dat                                / NODES_FILE
elements.dat                             / ELEMENTS_FILE
"""
        )

        config = read_preprocessor_main(main_file)
        # First data line at idx==0 is consumed as model_name
        assert config.model_name == "nodes.dat"
        # Since nodes.dat was consumed as model name, nodes_file is None
        assert config.nodes_file is None


class TestSaveModelToPreProcessorExtended:
    """Extended tests for save_model_to_preprocessor."""

    def test_save_model_without_mesh(self, tmp_path: Path) -> None:
        """Saving a model with no mesh still creates main file."""
        model = IWFMModel(name="EmptyModel")
        output_dir = tmp_path / "output_empty"
        config = save_model_to_preprocessor(model, output_dir)

        assert config.nodes_file is None
        assert config.elements_file is None
        main_file = output_dir / "EmptyModel_pp.in"
        assert main_file.exists()

    def test_save_model_without_stratigraphy(
        self,
        tmp_path: Path,
        small_grid_nodes: list[dict],
        small_grid_elements: list[dict],
    ) -> None:
        """Saving a model with mesh but no stratigraphy."""
        nodes = {d["id"]: Node(**d) for d in small_grid_nodes}
        elements = {d["id"]: Element(**d) for d in small_grid_elements}
        mesh = AppGrid(nodes=nodes, elements=elements)

        model = IWFMModel(name="NoStrat", mesh=mesh)
        output_dir = tmp_path / "output_nostrat"
        config = save_model_to_preprocessor(model, output_dir)

        assert config.nodes_file is not None
        assert config.nodes_file.exists()
        assert config.elements_file is not None
        assert config.elements_file.exists()
        assert config.stratigraphy_file is None

    def test_save_model_without_subregions(
        self,
        tmp_path: Path,
        small_grid_nodes: list[dict],
        small_grid_elements: list[dict],
    ) -> None:
        """Saving a model with mesh but no subregions defined."""
        nodes = {d["id"]: Node(**d) for d in small_grid_nodes}
        elements = {d["id"]: Element(**d) for d in small_grid_elements}
        mesh = AppGrid(nodes=nodes, elements=elements)
        # mesh has no subregions dict entries

        model = IWFMModel(name="NoSub", mesh=mesh)
        output_dir = tmp_path / "output_nosub"
        config = save_model_to_preprocessor(model, output_dir)

        # No subregions file should be created
        assert config.subregions_file is None

    def test_save_with_custom_name(
        self,
        tmp_path: Path,
        small_grid_nodes: list[dict],
        small_grid_elements: list[dict],
    ) -> None:
        """model_name parameter overrides model.name."""
        nodes = {d["id"]: Node(**d) for d in small_grid_nodes}
        elements = {d["id"]: Element(**d) for d in small_grid_elements}
        mesh = AppGrid(nodes=nodes, elements=elements)

        model = IWFMModel(name="OriginalName", mesh=mesh)
        output_dir = tmp_path / "output_custom"
        config = save_model_to_preprocessor(model, output_dir, model_name="CustomName")

        assert config.model_name == "CustomName"
        main_file = output_dir / "CustomName_pp.in"
        assert main_file.exists()

    def test_save_model_name_defaults(self, tmp_path: Path) -> None:
        """When model has no name and no override, defaults to 'iwfm_model'."""
        model = IWFMModel(name="")
        output_dir = tmp_path / "output_default"
        config = save_model_to_preprocessor(model, output_dir)

        assert config.model_name == "iwfm_model"


class TestPreProcessorConfigExtended:
    """Extended tests for PreProcessorConfig dataclass."""

    def test_default_values(self, tmp_path: Path) -> None:
        """Test all default values of PreProcessorConfig."""
        config = PreProcessorConfig(base_dir=tmp_path)

        assert config.model_name == ""
        assert config.nodes_file is None
        assert config.elements_file is None
        assert config.stratigraphy_file is None
        assert config.subregions_file is None
        assert config.streams_file is None
        assert config.lakes_file is None
        assert config.groundwater_file is None
        assert config.rootzone_file is None
        assert config.pumping_file is None
        assert config.output_dir is None
        assert config.budget_output_file is None
        assert config.heads_output_file is None
        assert config.n_layers == 1
        assert config.length_unit == "FT"
        assert config.area_unit == "ACRES"
        assert config.volume_unit == "AF"
        assert config.metadata == {}

    def test_metadata_dict(self, tmp_path: Path) -> None:
        """Test that metadata can be set and retrieved."""
        config = PreProcessorConfig(
            base_dir=tmp_path,
            metadata={"key1": "value1", "key2": 42},
        )
        assert config.metadata["key1"] == "value1"
        assert config.metadata["key2"] == 42

    def test_all_file_paths_set(self, tmp_path: Path) -> None:
        """Test setting all file path fields."""
        config = PreProcessorConfig(
            base_dir=tmp_path,
            nodes_file=tmp_path / "n.dat",
            elements_file=tmp_path / "e.dat",
            stratigraphy_file=tmp_path / "s.dat",
            subregions_file=tmp_path / "sr.dat",
            streams_file=tmp_path / "st.dat",
            lakes_file=tmp_path / "l.dat",
            groundwater_file=tmp_path / "gw.dat",
            rootzone_file=tmp_path / "rz.dat",
            pumping_file=tmp_path / "p.dat",
            output_dir=tmp_path / "out",
            budget_output_file=tmp_path / "budget.out",
            heads_output_file=tmp_path / "heads.out",
        )
        assert config.nodes_file is not None
        assert config.elements_file is not None
        assert config.stratigraphy_file is not None
        assert config.subregions_file is not None
        assert config.streams_file is not None
        assert config.lakes_file is not None
        assert config.groundwater_file is not None
        assert config.rootzone_file is not None
        assert config.pumping_file is not None
        assert config.output_dir is not None
        assert config.budget_output_file is not None
        assert config.heads_output_file is not None


class TestReadPreProcessorMainEdgeCases:
    """Edge case tests for read_preprocessor_main."""

    def test_first_data_line_without_name_desc(self, tmp_path: Path) -> None:
        """First data line is treated as model name even without NAME in desc."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  Header
SomeModelName
nodes.dat                                / NODES_FILE
"""
        )
        config = read_preprocessor_main(main_file)
        # First data line at idx==0 is captured as model_name
        assert config.model_name == "SomeModelName"

    def test_empty_file(self, tmp_path: Path) -> None:
        """Completely empty file returns default config."""
        main_file = tmp_path / "empty.in"
        main_file.write_text("")
        config = read_preprocessor_main(main_file)
        assert config.model_name == ""
        assert config.nodes_file is None

    def test_data_line_with_empty_value_skipped(self, tmp_path: Path) -> None:
        """Lines that parse to empty value are skipped."""
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  Header
Model                                    / MODEL_NAME
                                         / EMPTY_VALUE
nodes.dat                                / NODES_FILE
"""
        )
        config = read_preprocessor_main(main_file)
        assert config.model_name == "Model"
        assert config.nodes_file == tmp_path / "nodes.dat"

    def test_description_case_insensitive(self, tmp_path: Path) -> None:
        """Description keywords are matched case-insensitively (via .upper()).

        Note: values starting with 'c' (lowercase) are treated as
        comment lines by _is_comment_line, so 'cu_m' would be lost.
        We use 'AF' for volume_unit to avoid this issue.
        """
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  Header
Model                                    / model_name
nodes.dat                                / Nodes_File
elements.dat                             / Elements_File
2                                        / n_layers
m                                        / length_unit
sq_m                                     / area_unit
AF                                       / volume_unit
"""
        )
        config = read_preprocessor_main(main_file)
        assert config.model_name == "Model"
        assert config.nodes_file == tmp_path / "nodes.dat"
        assert config.elements_file == tmp_path / "elements.dat"
        assert config.n_layers == 2
        assert config.length_unit == "M"
        assert config.area_unit == "SQ_M"
        assert config.volume_unit == "AF"

    def test_value_starting_with_c_treated_as_comment(self, tmp_path: Path) -> None:
        """A data line whose value starts with C/c/* in column 1 is
        treated as a comment and the value is silently lost.
        """
        main_file = tmp_path / "test_pp.in"
        main_file.write_text(
            """C  Header
Model                                    / MODEL_NAME
cu_m                                     / VOLUME_UNIT
"""
        )
        config = read_preprocessor_main(main_file)
        # "cu_m / VOLUME_UNIT" starts with 'c' -> treated as comment
        assert config.volume_unit == "AF"  # default unchanged


class TestWritePreProcessorMainDefaultHeader:
    """Tests for default header generation in write_preprocessor_main."""

    def test_default_header_written(self, tmp_path: Path) -> None:
        """Default header includes 'IWFM PreProcessor' and 'pyiwfm'."""
        config = PreProcessorConfig(base_dir=tmp_path, model_name="Test")
        main_file = tmp_path / "test_pp.in"
        write_preprocessor_main(main_file, config)

        content = main_file.read_text()
        assert "IWFM PreProcessor Main Input File" in content
        assert "pyiwfm" in content


class TestWriteSubregionsRoundtrip:
    """Roundtrip tests for subregion writing and reading."""

    def test_single_subregion_roundtrip(self, tmp_path: Path) -> None:
        """Single subregion survives write-read roundtrip."""
        subregions = {1: Subregion(id=1, name="Only Region")}
        sr_file = tmp_path / "sr.dat"
        _write_subregions_file(sr_file, subregions)

        result = read_subregions_file(sr_file)
        assert len(result) == 1
        assert result[1].id == 1
        assert result[1].name == "Only Region"

    def test_subregions_written_in_sorted_order(self, tmp_path: Path) -> None:
        """Subregions are written in sorted order by ID."""
        subregions = {
            3: Subregion(id=3, name="Third"),
            1: Subregion(id=1, name="First"),
            2: Subregion(id=2, name="Second"),
        }
        sr_file = tmp_path / "sr.dat"
        _write_subregions_file(sr_file, subregions)

        content = sr_file.read_text()
        lines = [line for line in content.split("\n") if line.strip() and not line.startswith("C")]
        # First non-comment line is count, then sorted IDs
        data_lines = lines[1:]  # skip NSUBREGION line
        assert data_lines[0].strip().startswith("1")
        assert data_lines[1].strip().startswith("2")
        assert data_lines[2].strip().startswith("3")



"""
Lake component I/O handlers for IWFM model files.

This module provides functions for reading and writing IWFM lake
component files including lake definitions, lake elements, rating
curves, and outflows.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import TextIO

import numpy as np

from pyiwfm.components.lake import (
    AppLake,
    Lake,
    LakeElement,
    LakeOutflow,
    LakeRating,
)
from pyiwfm.core.exceptions import FileFormatError
from pyiwfm.io.iwfm_reader import (
    COMMENT_CHARS,
)
from pyiwfm.io.iwfm_reader import (
    is_comment_line as _is_comment_line,
)
from pyiwfm.io.iwfm_reader import (
    next_data_or_empty as _next_data_or_empty,
)
from pyiwfm.io.iwfm_reader import (
    resolve_path as _resolve_path_f,
)
from pyiwfm.io.iwfm_reader import (
    strip_inline_comment as _strip_comment,
)


@dataclass
class LakeFileConfig:
    """
    Configuration for lake component files.

    Attributes:
        output_dir: Directory for output files
        lakes_file: Lake definitions file name
        lake_elements_file: Lake elements file name
        rating_curves_file: Lake rating curves file name
        outflows_file: Lake outflows file name
    """

    output_dir: Path
    lakes_file: str = "lakes.dat"
    lake_elements_file: str = "lake_elements.dat"
    rating_curves_file: str = "lake_rating_curves.dat"
    outflows_file: str = "lake_outflows.dat"

    def get_lakes_path(self) -> Path:
        return self.output_dir / self.lakes_file

    def get_lake_elements_path(self) -> Path:
        return self.output_dir / self.lake_elements_file

    def get_rating_curves_path(self) -> Path:
        return self.output_dir / self.rating_curves_file

    def get_outflows_path(self) -> Path:
        return self.output_dir / self.outflows_file


class LakeWriter:
    """
    Writer for IWFM lake component files.

    Writes all lake-related input files including definitions, elements,
    rating curves, and outflows.

    Example:
        >>> config = LakeFileConfig(output_dir=Path("./model"))
        >>> writer = LakeWriter(config)
        >>> files = writer.write(lake_component)
    """

    def __init__(self, config: LakeFileConfig) -> None:
        """
        Initialize the lake writer.

        Args:
            config: File configuration
        """
        self.config = config
        config.output_dir.mkdir(parents=True, exist_ok=True)

    def write(self, lakes: AppLake) -> dict[str, Path]:
        """
        Write all lake component files.

        Args:
            lakes: AppLake component to write

        Returns:
            Dictionary mapping file type to output path
        """
        files: dict[str, Path] = {}

        # Write lakes definition
        if lakes.lakes:
            files["lakes"] = self.write_lake_definitions(lakes)

        # Write lake elements
        if lakes.lake_elements:
            files["lake_elements"] = self.write_lake_elements(lakes)

        # Write rating curves for lakes that have them
        lakes_with_ratings = [lk for lk in lakes.lakes.values() if lk.rating is not None]
        if lakes_with_ratings:
            files["rating_curves"] = self.write_rating_curves(lakes)

        # Write outflows for lakes that have them
        lakes_with_outflows = [lk for lk in lakes.lakes.values() if lk.outflow is not None]
        if lakes_with_outflows:
            files["outflows"] = self.write_outflows(lakes)

        return files

    def write_lake_definitions(self, lakes: AppLake, header: str | None = None) -> Path:
        """
        Write lake definitions file.

        Args:
            lakes: AppLake component
            header: Optional header comment

        Returns:
            Path to written file
        """
        filepath = self.config.get_lakes_path()

        with open(filepath, "w") as f:
            # Write header
            if header:
                for line in header.strip().split("\n"):
                    f.write(f"C  {line}\n")
            else:
                f.write("C  Lake definitions file\n")
                f.write("C  Generated by pyiwfm\n")
                f.write("C\n")
                f.write("C  ID   MAX_ELEV     INIT_STORAGE  NAME\n")

            # Write lake count
            f.write(f"{len(lakes.lakes):<10}                              / NLAKES\n")

            # Write lakes in ID order
            for lake_id in sorted(lakes.lakes.keys()):
                lake = lakes.lakes[lake_id]
                max_elev = lake.max_elevation if lake.max_elevation != float("inf") else 9999.0
                f.write(
                    f"{lake.id:<6} {max_elev:>12.4f} {lake.initial_storage:>14.4f}  {lake.name}\n"
                )

        return filepath

    def write_lake_elements(self, lakes: AppLake, header: str | None = None) -> Path:
        """
        Write lake elements file.

        Args:
            lakes: AppLake component
            header: Optional header comment

        Returns:
            Path to written file
        """
        filepath = self.config.get_lake_elements_path()

        with open(filepath, "w") as f:
            # Write header
            if header:
                for line in header.strip().split("\n"):
                    f.write(f"C  {line}\n")
            else:
                f.write("C  Lake elements file\n")
                f.write("C  Generated by pyiwfm\n")
                f.write("C\n")
                f.write("C  ELEM_ID  LAKE_ID  FRACTION\n")

            # Write element count
            f.write(
                f"{len(lakes.lake_elements):<10}                              / NLAKE_ELEMENTS\n"
            )

            # Write lake elements
            for elem in lakes.lake_elements:
                f.write(f"{elem.element_id:>7} {elem.lake_id:>7} {elem.fraction:>10.6f}\n")

        return filepath

    def write_rating_curves(self, lakes: AppLake, header: str | None = None) -> Path:
        """
        Write lake rating curves file.

        Args:
            lakes: AppLake component
            header: Optional header comment

        Returns:
            Path to written file
        """
        filepath = self.config.get_rating_curves_path()

        with open(filepath, "w") as f:
            # Write header
            if header:
                for line in header.strip().split("\n"):
                    f.write(f"C  {line}\n")
            else:
                f.write("C  Lake rating curves file\n")
                f.write("C  Generated by pyiwfm\n")
                f.write("C\n")

            # Count lakes with rating curves
            lakes_with_ratings = [lk for lk in lakes.lakes.values() if lk.rating is not None]
            f.write(
                f"{len(lakes_with_ratings):<10}                              / N_RATING_CURVES\n"
            )

            # Write each rating curve
            for lake in lakes_with_ratings:
                rating = lake.rating
                assert rating is not None
                f.write("C\n")
                f.write(f"C  Rating curve for lake {lake.id}: {lake.name}\n")
                f.write(f"{lake.id:<6} {len(rating.elevations):>5}  / LAKE_ID, N_POINTS\n")
                f.write("C  ELEVATION       AREA          VOLUME\n")

                for i in range(len(rating.elevations)):
                    f.write(
                        f"{rating.elevations[i]:>12.4f} {rating.areas[i]:>14.4f} "
                        f"{rating.volumes[i]:>14.4f}\n"
                    )

        return filepath

    def write_outflows(self, lakes: AppLake, header: str | None = None) -> Path:
        """
        Write lake outflows file.

        Args:
            lakes: AppLake component
            header: Optional header comment

        Returns:
            Path to written file
        """
        filepath = self.config.get_outflows_path()

        with open(filepath, "w") as f:
            # Write header
            if header:
                for line in header.strip().split("\n"):
                    f.write(f"C  {line}\n")
            else:
                f.write("C  Lake outflows file\n")
                f.write("C  Generated by pyiwfm\n")
                f.write("C\n")
                f.write("C  LAKE_ID  DEST_TYPE    DEST_ID    MAX_RATE\n")

            # Count lakes with outflows
            lakes_with_outflows = [lk for lk in lakes.lakes.values() if lk.outflow is not None]
            f.write(f"{len(lakes_with_outflows):<10}                              / N_OUTFLOWS\n")

            # Write outflows
            for lake in lakes_with_outflows:
                outflow = lake.outflow
                assert outflow is not None
                max_rate = outflow.max_rate if outflow.max_rate != float("inf") else 9999999.0
                f.write(
                    f"{lake.id:>7} {outflow.destination_type:<12} {outflow.destination_id:>7} "
                    f"{max_rate:>14.4f}\n"
                )

        return filepath


class LakeReader:
    """
    Reader for IWFM lake component files.
    """

    def read_lake_definitions(self, filepath: Path | str) -> dict[int, Lake]:
        """
        Read lake definitions from file.

        Args:
            filepath: Path to lakes file

        Returns:
            Dictionary mapping lake ID to Lake object
        """
        filepath = Path(filepath)
        lakes: dict[int, Lake] = {}

        with open(filepath) as f:
            line_num = 0
            n_lakes = None

            # Find NLAKES (skip version header lines starting with #)
            for line in f:
                line_num += 1
                if _is_comment_line(line):
                    continue
                stripped = line.strip()
                if stripped.startswith("#"):
                    # Version header (e.g., "#4.0") — skip
                    continue

                value, _ = _strip_comment(line)
                try:
                    n_lakes = int(value)
                except ValueError as e:
                    raise FileFormatError(
                        f"Invalid NLAKES value: '{value}'", line_number=line_num
                    ) from e
                break

            if n_lakes is None:
                raise FileFormatError("Could not find NLAKES in file")

            # Read lake data — detect preprocessor vs simulation format
            lakes_read = 0
            for line in f:
                line_num += 1
                if _is_comment_line(line):
                    continue

                parts = line.split()
                if len(parts) < 3:
                    continue

                try:
                    lake_id = int(parts[0])

                    # Preprocessor format: ID TYPDST DST NELAKE IELAKE(first)
                    # Simulation format:   ID MAX_ELEV INITIAL_STORAGE [NAME]
                    if len(parts) >= 5 and lakes_read == 0:
                        # Check if this is preprocessor format by testing
                        # if columns 1-3 are all integers
                        try:
                            typdst = int(parts[1])
                            dst = int(parts[2])
                            nelake = int(parts[3])
                            is_preproc = True
                        except ValueError:
                            is_preproc = False
                    else:
                        is_preproc = lakes_read == 0 and False

                    if is_preproc:
                        # Preprocessor Lake.dat format
                        first_elem = int(parts[4]) if len(parts) > 4 else 0
                        elem_ids = [first_elem] if first_elem > 0 else []

                        # Read remaining NELAKE-1 continuation element rows
                        for _ce in range(nelake - 1):
                            for cont_line in f:
                                line_num += 1
                                if _is_comment_line(cont_line):
                                    continue
                                cont_parts = cont_line.split()
                                if cont_parts:
                                    elem_ids.append(int(cont_parts[0]))
                                break

                        lake = Lake(
                            id=lake_id,
                            elements=elem_ids,
                        )
                        if typdst == 1:
                            lake.outflow = LakeOutflow(
                                lake_id=lake_id,
                                destination_type="stream",
                                destination_id=dst,
                            )
                        lakes[lake_id] = lake
                        lakes_read += 1

                        # Read remaining lakes in preprocessor format
                        for _lk in range(n_lakes - 1):
                            for lline in f:
                                line_num += 1
                                if _is_comment_line(lline):
                                    continue
                                lp = lline.split()
                                if len(lp) < 5:
                                    break
                                lid = int(lp[0])
                                lt = int(lp[1])
                                ld = int(lp[2])
                                nl = int(lp[3])
                                fe = int(lp[4]) if len(lp) > 4 else 0
                                eids = [fe] if fe > 0 else []
                                for _ce2 in range(nl - 1):
                                    for cl2 in f:
                                        line_num += 1
                                        if _is_comment_line(cl2):
                                            continue
                                        cp2 = cl2.split()
                                        if cp2:
                                            eids.append(int(cp2[0]))
                                        break
                                lk = Lake(id=lid, elements=eids)
                                if lt == 1:
                                    lk.outflow = LakeOutflow(
                                        lake_id=lid,
                                        destination_type="stream",
                                        destination_id=ld,
                                    )
                                lakes[lid] = lk
                                lakes_read += 1
                                break
                        break  # Done reading all lakes in preprocessor format
                    else:
                        # Simulation format: ID MAX_ELEV INITIAL_STORAGE [NAME]
                        max_elev = float(parts[1])
                        initial_storage = float(parts[2])
                        name = " ".join(parts[3:]) if len(parts) > 3 else ""

                        lakes[lake_id] = Lake(
                            id=lake_id,
                            max_elevation=max_elev if max_elev < 9990 else float("inf"),
                            initial_storage=initial_storage,
                            name=name,
                        )
                        lakes_read += 1

                except ValueError as e:
                    raise FileFormatError(
                        f"Invalid lake data: '{line.strip()}'", line_number=line_num
                    ) from e

        return lakes

    def read_lake_elements(self, filepath: Path | str) -> list[LakeElement]:
        """
        Read lake elements from file.

        Args:
            filepath: Path to lake elements file

        Returns:
            List of LakeElement objects
        """
        filepath = Path(filepath)
        elements: list[LakeElement] = []

        with open(filepath) as f:
            line_num = 0
            n_elements = None

            # Find NLAKE_ELEMENTS
            for line in f:
                line_num += 1
                if _is_comment_line(line):
                    continue

                value, _ = _strip_comment(line)
                try:
                    n_elements = int(value)
                except ValueError as e:
                    raise FileFormatError(
                        f"Invalid NLAKE_ELEMENTS value: '{value}'",
                        line_number=line_num,
                    ) from e
                break

            if n_elements is None:
                raise FileFormatError("Could not find NLAKE_ELEMENTS in file")

            # Read element data
            for line in f:
                line_num += 1
                if _is_comment_line(line):
                    continue

                parts = line.split()
                if len(parts) < 2:
                    continue

                try:
                    elem_id = int(parts[0])
                    lake_id = int(parts[1])
                    fraction = float(parts[2]) if len(parts) > 2 else 1.0

                    elements.append(
                        LakeElement(
                            element_id=elem_id,
                            lake_id=lake_id,
                            fraction=fraction,
                        )
                    )

                except ValueError as e:
                    raise FileFormatError(
                        f"Invalid lake element data: '{line.strip()}'",
                        line_number=line_num,
                    ) from e

        return elements

    def read_rating_curves(self, filepath: Path | str) -> dict[int, LakeRating]:
        """
        Read lake rating curves from file.

        Args:
            filepath: Path to rating curves file

        Returns:
            Dictionary mapping lake ID to LakeRating object
        """
        filepath = Path(filepath)
        ratings: dict[int, LakeRating] = {}

        with open(filepath) as f:
            line_num = 0
            n_curves = None

            # Find N_RATING_CURVES
            for line in f:
                line_num += 1
                if _is_comment_line(line):
                    continue

                value, _ = _strip_comment(line)
                try:
                    n_curves = int(value)
                except ValueError as e:
                    raise FileFormatError(
                        f"Invalid N_RATING_CURVES value: '{value}'",
                        line_number=line_num,
                    ) from e
                break

            if n_curves is None:
                raise FileFormatError("Could not find N_RATING_CURVES in file")

            # Read each rating curve
            curves_read = 0
            while curves_read < n_curves:
                # Find lake ID and n_points
                for line in f:
                    line_num += 1
                    if _is_comment_line(line):
                        continue

                    parts = line.split()
                    if len(parts) >= 2:
                        lake_id = int(parts[0])
                        n_points = int(parts[1])
                        break

                # Read rating data
                elevations = []
                areas = []
                volumes = []

                points_read = 0
                for line in f:
                    line_num += 1
                    if _is_comment_line(line):
                        continue

                    parts = line.split()
                    if len(parts) >= 3:
                        elevations.append(float(parts[0]))
                        areas.append(float(parts[1]))
                        volumes.append(float(parts[2]))
                        points_read += 1

                        if points_read >= n_points:
                            break

                ratings[lake_id] = LakeRating(
                    elevations=np.array(elevations),
                    areas=np.array(areas),
                    volumes=np.array(volumes),
                )

                curves_read += 1

        return ratings


# =============================================================================
# Lake Component Main File Reader (hierarchical dispatcher file)
# =============================================================================


@dataclass
class LakeParamSpec:
    """
    Per-lake parameters from the simulation lake main file.

    Attributes:
        lake_id: Lake ID
        conductance_coeff: Dimensionless conductance coefficient (CLAKE)
        depth_denom: Lake depth denominator (DLAKE)
        max_elev_col: Column index in max elevation file
        et_col: Column index in ET time series file
        precip_col: Column index in precipitation file
        name: Lake identification name
    """

    lake_id: int = 0
    conductance_coeff: float = 0.0
    depth_denom: float = 1.0
    max_elev_col: int = 0
    et_col: int = 0
    precip_col: int = 0
    name: str = ""


@dataclass
class OutflowRatingPoint:
    """A single point in a lake outflow rating table (v5.0).

    Attributes:
        elevation: Water surface elevation
        outflow: Outflow rate at this elevation
    """

    elevation: float = 0.0
    outflow: float = 0.0


@dataclass
class LakeOutflowRating:
    """Outflow rating table for a lake (v5.0 only).

    Attributes:
        lake_id: Lake ID
        points: List of (elevation, outflow) rating points
    """

    lake_id: int = 0
    points: list[OutflowRatingPoint] = field(default_factory=list)


@dataclass
class LakeMainFileConfig:
    """
    Configuration parsed from Lake component main simulation file.

    The lake main file configures lake hydrological parameters and
    data source file paths for the simulation.

    Attributes:
        version: File format version (e.g., "4.0" or "5.0")
        max_elev_file: Path to max lake elevation time series file
        budget_output_file: Path to lake budget output file
        final_elev_file: Path to end-of-simulation elevation output file
        conductance_factor: Conductance factor K (FactK)
        conductance_time_unit: Time unit for conductance
        depth_factor: Lake depth parameter L (FactL)
        lake_params: Per-lake parameter specifications
        elev_factor: Elevation conversion factor (v5.0 only)
        outflow_factor: Outflow rate conversion factor (v5.0 only)
        outflow_time_unit: Time unit for outflow rating tables (v5.0 only)
        outflow_ratings: Outflow rating tables per lake (v5.0 only)
    """

    version: str = ""
    max_elev_file: Path | None = None
    budget_output_file: Path | None = None
    final_elev_file: Path | None = None
    conductance_factor: float = 1.0
    conductance_time_unit: str = ""
    depth_factor: float = 1.0
    lake_params: list[LakeParamSpec] = field(default_factory=list)

    # v5.0 outflow rating tables
    elev_factor: float = 1.0
    outflow_factor: float = 1.0
    outflow_time_unit: str = ""
    outflow_ratings: list[LakeOutflowRating] = field(default_factory=list)


class LakeMainFileReader:
    """
    Reader for IWFM lake component main simulation file.

    The Lake main file is a hierarchical dispatcher that contains:
    1. Version header (e.g., #4.0)
    2. Paths to sub-files (max elevation TS, budget output, final elevation)
    3. Conductance and depth parameters
    4. Per-lake parameter lines
    5. (v5.0) Outflow rating tables

    Supports both version 4.0 and 5.0 formats.
    """

    def __init__(self) -> None:
        self._line_num = 0

    def read(self, filepath: Path | str, base_dir: Path | None = None) -> LakeMainFileConfig:
        """
        Parse Lake main file.

        Args:
            filepath: Path to the Lake component main file
            base_dir: Base directory for resolving relative paths.
                     If None, uses the parent directory of filepath.

        Returns:
            LakeMainFileConfig with parsed values
        """
        filepath = Path(filepath)
        if base_dir is None:
            base_dir = filepath.parent

        config = LakeMainFileConfig()
        self._line_num = 0

        with open(filepath) as f:
            # Read version header
            config.version = self._read_version(f)
            is_v50 = config.version.startswith("5")

            # Max lake elevation file path
            max_elev_path = _next_data_or_empty(f)
            if max_elev_path:
                config.max_elev_file = _resolve_path_f(base_dir, max_elev_path)

            # Budget output file path
            budget_path = _next_data_or_empty(f)
            if budget_path:
                config.budget_output_file = _resolve_path_f(base_dir, budget_path)

            # Final elevation output file path (optional)
            final_path = _next_data_or_empty(f)
            if final_path:
                config.final_elev_file = _resolve_path_f(base_dir, final_path)

            # Conductance factor K
            factk_str = _next_data_or_empty(f)
            if factk_str:
                config.conductance_factor = float(factk_str)

            # Conductance time unit
            config.conductance_time_unit = _next_data_or_empty(f)

            # Depth factor L
            factl_str = _next_data_or_empty(f)
            if factl_str:
                config.depth_factor = float(factl_str)

            # Per-lake parameter lines
            # We read lines until we encounter something that doesn't parse
            # as a lake parameter line. Each line has:
            # CLAKE_ID  CLAKE  DLAKE  MaxElev_Col  ET_Col  Precip_Col  [Name]
            while True:
                data = _next_data_or_empty(f)
                if not data:
                    break

                parts = data.split()
                if len(parts) < 6:
                    # Not a lake param line - might be v5.0 elevation factor
                    # or initial conditions section
                    if is_v50 and len(parts) == 1:
                        try:
                            config.elev_factor = float(parts[0])
                            break
                        except ValueError:
                            break
                    break

                try:
                    param = LakeParamSpec(
                        lake_id=int(float(parts[0])),
                        conductance_coeff=float(parts[1]),
                        depth_denom=float(parts[2]),
                        max_elev_col=int(float(parts[3])),
                        et_col=int(float(parts[4])),
                        precip_col=int(float(parts[5])),
                        name=" ".join(parts[6:]) if len(parts) > 6 else "",
                    )
                    config.lake_params.append(param)
                except (ValueError, IndexError):
                    break

            # v5.0: Read outflow rating tables
            if is_v50 and config.lake_params:
                # Outflow rate factor (already read elev_factor above)
                factq_str = _next_data_or_empty(f)
                if factq_str:
                    config.outflow_factor = float(factq_str)

                # Outflow time unit
                config.outflow_time_unit = _next_data_or_empty(f)

                # Read outflow rating table for each lake
                for _ in range(len(config.lake_params)):
                    rating = self._read_outflow_rating(f, config.elev_factor, config.outflow_factor)
                    if rating:
                        config.outflow_ratings.append(rating)

        return config

    def _read_outflow_rating(
        self,
        f: TextIO,
        elev_factor: float,
        outflow_factor: float,
    ) -> LakeOutflowRating | None:
        """Read a single outflow rating table for a lake (v5.0).

        Format:
            Lake_ID  NPoints  Elev_1  Outflow_1
            Elev_2  Outflow_2
            ...
        """
        header = _next_data_or_empty(f)
        if not header:
            return None

        parts = header.split()
        if len(parts) < 4:
            return None

        rating = LakeOutflowRating(lake_id=int(float(parts[0])))
        n_points = int(float(parts[1]))

        # First point on header line
        rating.points.append(
            OutflowRatingPoint(
                elevation=float(parts[2]) * elev_factor,
                outflow=float(parts[3]) * outflow_factor,
            )
        )

        # Remaining points
        for _ in range(n_points - 1):
            line = _next_data_or_empty(f)
            if not line:
                break
            lp = line.split()
            if len(lp) >= 2:
                rating.points.append(
                    OutflowRatingPoint(
                        elevation=float(lp[0]) * elev_factor,
                        outflow=float(lp[1]) * outflow_factor,
                    )
                )

        return rating

    def _read_version(self, f: TextIO) -> str:
        """Read the version header from the file."""
        for line in f:
            self._line_num += 1
            stripped = line.strip()
            if not stripped:
                continue
            if stripped.startswith("#"):
                return stripped[1:].strip()
            if line[0] in COMMENT_CHARS:
                continue
            break
        return ""


# Convenience functions


def read_lake_main_file(filepath: Path | str, base_dir: Path | None = None) -> LakeMainFileConfig:
    """
    Read IWFM lake component main simulation file.

    Args:
        filepath: Path to the Lake component main file
        base_dir: Base directory for resolving relative paths.

    Returns:
        LakeMainFileConfig with parsed values
    """
    reader = LakeMainFileReader()
    return reader.read(filepath, base_dir)


def write_lakes(
    lakes: AppLake,
    output_dir: Path | str,
    config: LakeFileConfig | None = None,
) -> dict[str, Path]:
    """
    Write lake component to files.

    Args:
        lakes: AppLake component to write
        output_dir: Output directory
        config: Optional file configuration

    Returns:
        Dictionary mapping file type to output path
    """
    output_dir = Path(output_dir)

    if config is None:
        config = LakeFileConfig(output_dir=output_dir)
    else:
        config.output_dir = output_dir

    writer = LakeWriter(config)
    return writer.write(lakes)


def read_lake_definitions(filepath: Path | str) -> dict[int, Lake]:
    """
    Read lake definitions from file.

    Args:
        filepath: Path to lakes file

    Returns:
        Dictionary mapping lake ID to Lake object
    """
    reader = LakeReader()
    return reader.read_lake_definitions(filepath)


def read_lake_elements(filepath: Path | str) -> list[LakeElement]:
    """
    Read lake elements from file.

    Args:
        filepath: Path to lake elements file

    Returns:
        List of LakeElement objects
    """
    reader = LakeReader()
    return reader.read_lake_elements(filepath)



"""Tests for PEST++ integration module."""

from __future__ import annotations

from datetime import datetime
from pathlib import Path

import pytest

from pyiwfm.runner.pest import (
    InstructionFile,
    Observation,
    ObservationGroup,
    Parameter,
    PESTInterface,
    TemplateFile,
    write_pest_control_file,
)


class TestParameter:
    """Tests for Parameter dataclass."""

    def test_basic_creation(self):
        """Test creating a basic parameter."""
        param = Parameter(
            name="hk_zone1",
            initial_value=10.0,
            lower_bound=0.1,
            upper_bound=1000.0,
        )
        assert param.name == "hk_zone1"
        assert param.initial_value == 10.0
        assert param.lower_bound == 0.1
        assert param.upper_bound == 1000.0
        assert param.group == "default"
        assert param.transform == "none"

    def test_with_transform(self):
        """Test parameter with log transform."""
        param = Parameter(
            name="hk",
            initial_value=10.0,
            lower_bound=0.1,
            upper_bound=1000.0,
            transform="log",
        )
        assert param.transform == "log"

    def test_with_group(self):
        """Test parameter with custom group."""
        param = Parameter(
            name="hk",
            initial_value=10.0,
            lower_bound=0.1,
            upper_bound=1000.0,
            group="hydraulic_conductivity",
        )
        assert param.group == "hydraulic_conductivity"

    def test_invalid_name_length(self):
        """Test that long names raise error."""
        with pytest.raises(ValueError, match="name too long"):
            Parameter(
                name="x" * 201,  # Over 200 char limit
                initial_value=1.0,
                lower_bound=0.0,
                upper_bound=2.0,
            )

    def test_invalid_bounds(self):
        """Test that invalid bounds raise error."""
        with pytest.raises(ValueError, match="Lower bound"):
            Parameter(
                name="test",
                initial_value=1.0,
                lower_bound=10.0,  # Greater than upper
                upper_bound=5.0,
            )

    def test_initial_outside_bounds(self):
        """Test that initial value outside bounds raises error."""
        with pytest.raises(ValueError, match="Initial value"):
            Parameter(
                name="test",
                initial_value=100.0,  # Outside bounds
                lower_bound=0.0,
                upper_bound=10.0,
            )

    def test_to_pest_line(self):
        """Test formatting as PEST control file line."""
        param = Parameter(
            name="hk_zone1",
            initial_value=10.0,
            lower_bound=0.1,
            upper_bound=1000.0,
            group="hk",
        )
        line = param.to_pest_line()

        assert "hk_zone1" in line
        assert "none" in line
        assert "hk" in line


class TestObservation:
    """Tests for Observation dataclass."""

    def test_basic_creation(self):
        """Test creating a basic observation."""
        obs = Observation(
            name="head_well1_t100",
            value=-50.5,
            weight=1.0,
        )
        assert obs.name == "head_well1_t100"
        assert obs.value == -50.5
        assert obs.weight == 1.0
        assert obs.group == "default"

    def test_with_group(self):
        """Test observation with custom group."""
        obs = Observation(
            name="head_well1",
            value=-50.0,
            group="heads",
        )
        assert obs.group == "heads"

    def test_invalid_name_length(self):
        """Test that long names raise error."""
        with pytest.raises(ValueError, match="name too long"):
            Observation(name="x" * 201, value=0.0)

    def test_negative_weight(self):
        """Test that negative weight raises error."""
        with pytest.raises(ValueError, match="Weight must be non-negative"):
            Observation(name="test", value=0.0, weight=-1.0)

    def test_to_pest_line(self):
        """Test formatting as PEST control file line."""
        obs = Observation(
            name="head_well1",
            value=-50.5,
            weight=2.5,
            group="heads",
        )
        line = obs.to_pest_line()

        assert "head_well1" in line
        assert "heads" in line


class TestObservationGroup:
    """Tests for ObservationGroup dataclass."""

    def test_basic_creation(self):
        """Test creating a basic observation group."""
        group = ObservationGroup(name="heads")
        assert group.name == "heads"
        assert group.observations == []

    def test_add_observation(self):
        """Test adding observations to group."""
        group = ObservationGroup(name="heads")

        obs = group.add_observation("head_1", -50.0, 1.0)

        assert len(group.observations) == 1
        assert obs.name == "head_1"
        assert obs.group == "heads"

    def test_multiple_observations(self):
        """Test adding multiple observations."""
        group = ObservationGroup(name="flows")

        group.add_observation("flow_1", 100.0, 0.5)
        group.add_observation("flow_2", 200.0, 0.5)
        group.add_observation("flow_3", 150.0, 0.5)

        assert len(group.observations) == 3


class TestTemplateFile:
    """Tests for TemplateFile dataclass."""

    def test_basic_creation(self):
        """Test creating a basic template file."""
        tpl = TemplateFile(
            template_path=Path("model.tpl"),
            input_path=Path("model.in"),
        )
        assert tpl.template_path == Path("model.tpl")
        assert tpl.input_path == Path("model.in")
        assert tpl.delimiter == "#"
        assert tpl.parameters == []

    def test_path_conversion(self):
        """Test that string paths are converted."""
        tpl = TemplateFile(
            template_path="model.tpl",
            input_path="model.in",
        )
        assert isinstance(tpl.template_path, Path)
        assert isinstance(tpl.input_path, Path)

    def test_create_from_file(self, tmp_path):
        """Test creating template from existing input file."""
        # Create input file with parameter value
        input_file = tmp_path / "model.in"
        input_file.write_text("C Model input\nHK = 1.5e+01\nSS = 1.0e-05\n")

        template_file = tmp_path / "model.tpl"

        TemplateFile.create_from_file(
            input_file,
            template_file,
            parameters={"hk": 1.5e01},
        )

        assert template_file.exists()
        content = template_file.read_text()
        assert "ptf #" in content
        assert "#" in content  # Parameter marker

    def test_to_pest_line(self):
        """Test formatting as PEST control file line."""
        tpl = TemplateFile(
            template_path=Path("model.tpl"),
            input_path=Path("model.in"),
        )
        line = tpl.to_pest_line()

        assert "model.tpl" in line
        assert "model.in" in line


class TestInstructionFile:
    """Tests for InstructionFile dataclass."""

    def test_basic_creation(self):
        """Test creating a basic instruction file."""
        ins = InstructionFile(
            instruction_path=Path("output.ins"),
            output_path=Path("output.out"),
        )
        assert ins.instruction_path == Path("output.ins")
        assert ins.output_path == Path("output.out")
        assert ins.marker == "@"

    def test_create_for_timeseries(self, tmp_path):
        """Test creating instruction file for time series."""
        output_file = tmp_path / "hydrograph.out"
        instruction_file = tmp_path / "hydrograph.ins"

        ins = InstructionFile.create_for_timeseries(
            output_file,
            instruction_file,
            observations=[
                ("head_t1", 1, 2),  # Line 1, column 2
                ("head_t2", 2, 2),  # Line 2, column 2
            ],
            header_lines=1,
        )

        assert instruction_file.exists()
        assert len(ins.observations) == 2
        assert "head_t1" in ins.observations
        assert "head_t2" in ins.observations

        content = instruction_file.read_text()
        assert "pif @" in content

    def test_create_for_hydrograph(self, tmp_path):
        """Test creating instruction file for hydrograph."""
        output_file = tmp_path / "well_hydrograph.out"
        instruction_file = tmp_path / "well.ins"

        ins = InstructionFile.create_for_hydrograph(
            output_file,
            instruction_file,
            location_name="well1",
            observation_times=[
                (datetime(2000, 10, 1), "oct2000"),
                (datetime(2000, 11, 1), "nov2000"),
            ],
        )

        assert instruction_file.exists()
        assert "well1_oct2000" in ins.observations
        assert "well1_nov2000" in ins.observations

    def test_to_pest_line(self):
        """Test formatting as PEST control file line."""
        ins = InstructionFile(
            instruction_path=Path("output.ins"),
            output_path=Path("output.out"),
        )
        line = ins.to_pest_line()

        assert "output.ins" in line
        assert "output.out" in line


class TestPESTInterface:
    """Tests for PESTInterface class."""

    @pytest.fixture
    def model_dir(self, tmp_path):
        """Create a mock model directory."""
        model = tmp_path / "model"
        model.mkdir()
        (model / "Simulation.in").write_text("C Main file\n")
        return model

    def test_init(self, model_dir):
        """Test interface initialization."""
        pest = PESTInterface(
            model_dir=model_dir,
            case_name="test_cal",
        )

        assert pest.model_dir == model_dir
        assert pest.case_name == "test_cal"
        assert pest.pest_dir.exists()
        assert pest.parameters == []
        assert pest.observations == []

    def test_add_parameter(self, model_dir):
        """Test adding parameters."""
        pest = PESTInterface(model_dir=model_dir, case_name="test")

        param = pest.add_parameter(
            "hk_zone1",
            initial_value=10.0,
            lower_bound=0.1,
            upper_bound=1000.0,
            group="hk",
        )

        assert len(pest.parameters) == 1
        assert param.name == "hk_zone1"
        assert "hk" in pest.parameter_groups

    def test_add_observation(self, model_dir):
        """Test adding observations."""
        pest = PESTInterface(model_dir=model_dir, case_name="test")

        obs = pest.add_observation(
            "head_well1",
            value=-50.0,
            weight=1.0,
            group="heads",
        )

        assert len(pest.observations) == 1
        assert obs.name == "head_well1"
        assert "heads" in pest.observation_groups

    def test_add_observation_group(self, model_dir):
        """Test adding observation groups."""
        pest = PESTInterface(model_dir=model_dir, case_name="test")

        group = pest.add_observation_group(
            "heads",
            observations=[
                ("head_1", -50.0, 1.0),
                ("head_2", -45.0, 1.0),
            ],
        )

        assert group.name == "heads"
        assert len(pest.observations) == 2

    def test_set_model_command(self, model_dir):
        """Test setting model command."""
        pest = PESTInterface(model_dir=model_dir, case_name="test")

        pest.set_model_command("python run_iwfm.py")

        assert pest.model_command == "python run_iwfm.py"

    def test_set_pestpp_option(self, model_dir):
        """Test setting PEST++ options."""
        pest = PESTInterface(model_dir=model_dir, case_name="test")

        pest.set_pestpp_option("svd_pack", "redsvd")
        pest.set_pestpp_option("ies_num_reals", 100)

        assert pest.pestpp_options["svd_pack"] == "redsvd"
        assert pest.pestpp_options["ies_num_reals"] == 100

    def test_write_control_file(self, model_dir):
        """Test writing PEST++ control file."""
        pest = PESTInterface(model_dir=model_dir, case_name="test")

        # Add some parameters and observations
        pest.add_parameter("hk", 10.0, 0.1, 1000.0, group="hk")
        pest.add_observation("head_1", -50.0, 1.0, group="heads")

        # Add template and instruction files
        tpl = TemplateFile(
            template_path=pest.pest_dir / "model.tpl",
            input_path=model_dir / "model.in",
        )
        pest.add_template_file(tpl)

        ins = InstructionFile(
            instruction_path=pest.pest_dir / "output.ins",
            output_path=model_dir / "output.out",
        )
        pest.add_instruction_file(ins)

        # Write control file
        pst_file = pest.write_control_file()

        assert pst_file.exists()
        content = pst_file.read_text()

        assert "pcf" in content
        assert "* control data" in content
        assert "* parameter groups" in content
        assert "* parameter data" in content
        assert "* observation groups" in content
        assert "* observation data" in content
        assert "* model command line" in content
        assert "* model input/output" in content
        assert "hk" in content
        assert "head_1" in content

    def test_write_control_file_with_pestpp_options(self, model_dir):
        """Test control file includes PEST++ options."""
        pest = PESTInterface(model_dir=model_dir, case_name="test")

        pest.add_parameter("p1", 1.0, 0.1, 10.0)
        pest.add_observation("o1", 5.0)
        pest.set_pestpp_option("svd_pack", "redsvd")

        # Add minimal template and instruction
        tpl = TemplateFile(pest.pest_dir / "t.tpl", model_dir / "t.in")
        ins = InstructionFile(pest.pest_dir / "i.ins", model_dir / "o.out")
        pest.add_template_file(tpl)
        pest.add_instruction_file(ins)

        pst_file = pest.write_control_file()

        content = pst_file.read_text()
        assert "++" in content
        assert "svd_pack" in content

    def test_write_model_runner(self, model_dir):
        """Test writing model runner script."""
        pest = PESTInterface(model_dir=model_dir, case_name="test")

        script_path = pest.write_model_runner()

        assert script_path.exists()
        content = script_path.read_text()

        assert "#!/usr/bin/env python" in content
        assert "IWFMRunner" in content
        assert "run_simulation" in content

    def test_repr(self, model_dir):
        """Test string representation."""
        pest = PESTInterface(model_dir=model_dir, case_name="test")
        pest.add_parameter("p1", 1.0, 0.1, 10.0)
        pest.add_observation("o1", 5.0)

        repr_str = repr(pest)
        assert "PESTInterface" in repr_str
        assert "test" in repr_str
        assert "n_parameters=1" in repr_str
        assert "n_observations=1" in repr_str


class TestWritePestControlFile:
    """Tests for write_pest_control_file convenience function."""

    def test_write_control_file(self, tmp_path):
        """Test writing control file with convenience function."""
        pst_path = tmp_path / "test.pst"

        parameters = [
            Parameter("p1", 1.0, 0.1, 10.0, group="pgroup"),
        ]
        observations = [
            Observation("o1", 5.0, 1.0, group="ogroup"),
        ]
        template_files = [
            TemplateFile(tmp_path / "t.tpl", tmp_path / "t.in"),
        ]
        instruction_files = [
            InstructionFile(tmp_path / "i.ins", tmp_path / "o.out"),
        ]

        result = write_pest_control_file(
            pst_path,
            parameters,
            observations,
            template_files,
            instruction_files,
            svd_pack="redsvd",
        )

        assert result == pst_path
        assert pst_path.exists()

        content = pst_path.read_text()
        assert "p1" in content
        assert "o1" in content
        assert "svd_pack" in content



"""
Data export API routes: CSV, GeoJSON, GeoPackage, and plot downloads.
"""

from __future__ import annotations

import csv
import io
import json
import logging
import tempfile

from fastapi import APIRouter, HTTPException, Query
from fastapi.responses import Response

from pyiwfm.visualization.webapi.config import model_state

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/api/export", tags=["export"])


@router.get("/heads-csv")
def export_heads_csv(
    timestep: int = Query(default=0, ge=0, description="Timestep index"),
    layer: int = Query(default=1, ge=1, description="Layer number (1-based)"),
) -> Response:
    """
    Export head values as a CSV file.

    Returns per-node head values for the specified timestep and layer.
    """
    loader = model_state.get_head_loader()
    if loader is None:
        raise HTTPException(status_code=404, detail="No head data available")

    if timestep >= loader.n_frames:
        raise HTTPException(
            status_code=400,
            detail=f"Timestep {timestep} out of range [0, {loader.n_frames})",
        )

    frame = loader.get_frame(timestep)
    layer_idx = layer - 1
    if layer_idx >= frame.shape[1]:
        raise HTTPException(
            status_code=400,
            detail=f"Layer {layer} out of range [1, {frame.shape[1]}]",
        )

    values = frame[:, layer_idx]
    dt = loader.times[timestep] if timestep < len(loader.times) else None

    output = io.StringIO()
    writer = csv.writer(output)
    writer.writerow(["node_id", "head_ft"])
    for i, val in enumerate(values):
        writer.writerow([i + 1, round(float(val), 3)])

    filename = f"heads_ts{timestep}_layer{layer}"
    if dt:
        filename += f"_{dt.strftime('%Y%m%d')}"
    filename += ".csv"

    return Response(
        content=output.getvalue(),
        media_type="text/csv",
        headers={"Content-Disposition": f"attachment; filename={filename}"},
    )


@router.get("/mesh-geojson")
def export_mesh_geojson(
    layer: int = Query(default=1, ge=1, description="Layer number (1-based)"),
) -> Response:
    """
    Export the mesh as a GeoJSON file.

    Returns element polygons in WGS84 as a downloadable GeoJSON file.
    """
    if not model_state.is_loaded:
        raise HTTPException(status_code=404, detail="No model loaded")

    from pyiwfm.visualization.webapi.routes.mesh import get_mesh_geojson

    try:
        geojson = get_mesh_geojson(layer=layer)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e)) from e

    return Response(
        content=json.dumps(geojson),
        media_type="application/geo+json",
        headers={"Content-Disposition": f"attachment; filename=mesh_layer{layer}.geojson"},
    )


@router.get("/budget-csv")
def export_budget_csv(
    budget_type: str = Query(..., description="Budget type"),
    location: str = Query(default="", description="Location name or index"),
) -> Response:
    """
    Export budget time series data as a CSV file.
    """
    reader = model_state.get_budget_reader(budget_type)
    if reader is None:
        raise HTTPException(
            status_code=404,
            detail=f"Budget type '{budget_type}' not available",
        )

    loc = location if location else 0

    try:
        times_arr, values_arr = reader.get_values(loc)
        headers = reader.get_column_headers(loc)
    except (KeyError, IndexError) as e:
        raise HTTPException(status_code=400, detail=str(e)) from e

    # Build time strings
    ts = reader.header.timestep
    from datetime import timedelta

    use_months = "MON" in ts.unit.upper() if ts.unit else False
    if use_months:
        from dateutil.relativedelta import relativedelta

    time_strings = []
    if ts.start_datetime:
        for i in range(len(times_arr)):
            if use_months:
                dt = ts.start_datetime + relativedelta(months=i)
            else:
                dt = ts.start_datetime + timedelta(minutes=ts.delta_t_minutes * i)
            time_strings.append(dt.isoformat())
    else:
        time_strings = [str(t) for t in times_arr.tolist()]

    output = io.StringIO()
    writer = csv.writer(output)
    writer.writerow(["datetime"] + headers)
    for i in range(len(time_strings)):
        row = [time_strings[i]] + [
            round(float(values_arr[i, j]), 4) for j in range(values_arr.shape[1])
        ]
        writer.writerow(row)

    loc_name = location or reader.locations[0]
    safe_name = loc_name.replace(" ", "_").replace("/", "_")
    filename = f"budget_{budget_type}_{safe_name}.csv"

    return Response(
        content=output.getvalue(),
        media_type="text/csv",
        headers={"Content-Disposition": f"attachment; filename={filename}"},
    )


@router.get("/budget-excel")
def export_budget_excel(
    budget_type: str = Query(..., description="Budget type"),
    location: str = Query(default="", description="Location name/index or empty for all"),
) -> Response:
    """Export budget data as an Excel workbook."""
    reader = model_state.get_budget_reader(budget_type)
    if reader is None:
        raise HTTPException(
            status_code=404,
            detail=f"Budget type '{budget_type}' not available",
        )

    from pyiwfm.io.budget_excel import budget_to_excel

    location_ids: list[int] | None = None
    if location:
        try:
            loc_idx = reader.get_location_index(location)
            location_ids = [loc_idx + 1]
        except (KeyError, IndexError) as e:
            raise HTTPException(status_code=400, detail=str(e)) from e

    with tempfile.NamedTemporaryFile(suffix=".xlsx", delete=False) as tmp:
        tmp_path = tmp.name

    try:
        from pathlib import Path

        budget_to_excel(
            reader=reader,
            output_path=tmp_path,
            location_ids=location_ids,
        )
        data = Path(tmp_path).read_bytes()
    except Exception as e:
        logger.exception("Budget Excel export failed")
        raise HTTPException(status_code=500, detail=str(e)) from e
    finally:
        import os

        try:
            os.unlink(tmp_path)
        except OSError:
            pass

    safe_type = budget_type.replace(" ", "_").replace("/", "_")
    filename = f"budget_{safe_type}.xlsx"

    return Response(
        content=data,
        media_type="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        headers={"Content-Disposition": f"attachment; filename={filename}"},
    )


@router.get("/hydrograph-csv")
def export_hydrograph_csv(
    type: str = Query(..., description="Type: gw, stream, subsidence, tile_drain"),
    location_id: int = Query(..., description="Location/node ID"),
) -> Response:
    """
    Export hydrograph time series as a CSV file.
    """
    if not model_state.is_loaded:
        raise HTTPException(status_code=404, detail="No model loaded")

    if type == "gw":
        reader = model_state.get_gw_hydrograph_reader()
        if reader is None or reader.n_timesteps == 0:
            raise HTTPException(status_code=404, detail="No GW hydrograph data available")

        phys_locs = model_state.get_gw_physical_locations()
        location_index = location_id - 1

        if phys_locs:
            if location_index < 0 or location_index >= len(phys_locs):
                raise HTTPException(
                    status_code=404,
                    detail=f"GW hydrograph {location_id} out of range",
                )
            col_idx = phys_locs[location_index]["columns"][0][0]
        else:
            # Fallback: raw column index
            col_idx = location_index

        if col_idx < 0 or col_idx >= reader.n_columns:
            raise HTTPException(
                status_code=404,
                detail=f"GW hydrograph {location_id} out of range",
            )

        times, values = reader.get_time_series(col_idx)
        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerow(["datetime", "head_ft"])
        for t, v in zip(times, values, strict=False):
            writer.writerow([t, round(v, 3)])

        filename = f"hydrograph_gw_{location_id}.csv"

    elif type == "stream":
        reader = model_state.get_stream_hydrograph_reader()
        if reader is None or reader.n_timesteps == 0:
            raise HTTPException(
                status_code=404,
                detail="No stream hydrograph data available",
            )

        col_idx = reader.find_column_by_node_id(location_id)
        if col_idx is None and location_id in reader.hydrograph_ids:
            col_idx = reader.hydrograph_ids.index(location_id)
        if col_idx is None:
            raise HTTPException(
                status_code=404,
                detail=f"Stream node {location_id} not found",
            )

        times, values = reader.get_time_series(col_idx)
        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerow(["datetime", "flow_cfs"])
        for t, v in zip(times, values, strict=False):
            writer.writerow([t, round(v, 3)])

        filename = f"hydrograph_stream_{location_id}.csv"

    elif type == "subsidence":
        reader = model_state.get_subsidence_reader()
        if reader is None or reader.n_timesteps == 0:
            raise HTTPException(
                status_code=404,
                detail="No subsidence hydrograph data available",
            )

        col_idx = location_id - 1
        if col_idx < 0 or col_idx >= reader.n_columns:
            raise HTTPException(
                status_code=404,
                detail=f"Subsidence location {location_id} out of range",
            )

        times, values = reader.get_time_series(col_idx)
        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerow(["datetime", "subsidence_ft"])
        for t, v in zip(times, values, strict=False):
            writer.writerow([t, round(v, 3)])

        filename = f"hydrograph_subsidence_{location_id}.csv"

    elif type == "tile_drain":
        reader = model_state.get_tile_drain_reader()
        if reader is None or reader.n_timesteps == 0:
            raise HTTPException(
                status_code=404,
                detail="No tile drain hydrograph data available",
            )

        col_idx = location_id - 1
        if col_idx < 0 or col_idx >= reader.n_columns:
            raise HTTPException(
                status_code=404,
                detail=f"Tile drain location {location_id} out of range",
            )

        times, values = reader.get_time_series(col_idx)
        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerow(["datetime", "flow_volume"])
        for t, v in zip(times, values, strict=False):
            writer.writerow([t, round(v, 3)])

        filename = f"hydrograph_tile_drain_{location_id}.csv"

    else:
        raise HTTPException(
            status_code=400,
            detail=f"Unknown type: {type}. Use: gw, stream, subsidence, tile_drain",
        )

    return Response(
        content=output.getvalue(),
        media_type="text/csv",
        headers={"Content-Disposition": f"attachment; filename={filename}"},
    )


@router.get("/geopackage")
def export_geopackage(
    include_streams: bool = Query(default=True, description="Include stream reaches"),
    include_subregions: bool = Query(default=True, description="Include subregion polygons"),
    include_boundary: bool = Query(default=True, description="Include model boundary"),
) -> Response:
    """Export the model mesh as a GeoPackage file.

    Creates a multi-layer GeoPackage containing nodes, elements,
    and optionally streams, subregions, and boundary polygon.
    """
    if not model_state.is_loaded:
        raise HTTPException(status_code=404, detail="No model loaded")

    model = model_state.model
    if model is None or model.grid is None:
        raise HTTPException(status_code=404, detail="No mesh/grid loaded")

    from pyiwfm.visualization.gis_export import GISExporter

    exporter = GISExporter(
        grid=model.grid,
        stratigraphy=model.stratigraphy,
        streams=model.streams,
        crs=model_state._crs,
    )

    with tempfile.NamedTemporaryFile(suffix=".gpkg", delete=False) as tmp:
        tmp_path = tmp.name

    try:
        exporter.export_geopackage(
            tmp_path,
            include_streams=include_streams,
            include_subregions=include_subregions,
            include_boundary=include_boundary,
        )

        from pathlib import Path

        data = Path(tmp_path).read_bytes()

        model_name = model.name or "model"
        safe_name = model_name.replace(" ", "_").replace("/", "_")
        filename = f"{safe_name}.gpkg"

        return Response(
            content=data,
            media_type="application/geopackage+sqlite3",
            headers={"Content-Disposition": f"attachment; filename={filename}"},
        )
    except Exception as e:
        logger.exception("GeoPackage export failed")
        raise HTTPException(status_code=500, detail=str(e)) from e
    finally:
        import os

        try:
            os.unlink(tmp_path)
        except OSError:
            pass


@router.get("/plot/{plot_type}")
def export_plot(
    plot_type: str,
    format: str = Query(default="png", description="Image format: png or svg"),
    layer: int = Query(default=1, ge=1, description="Layer number (1-based)"),
    timestep: int = Query(default=0, ge=0, description="Timestep index"),
    width: float = Query(default=10.0, gt=0, description="Figure width in inches"),
    height: float = Query(default=8.0, gt=0, description="Figure height in inches"),
    dpi: int = Query(default=150, ge=72, le=600, description="DPI for PNG output"),
) -> Response:
    """Generate publication-quality matplotlib figures.

    Supported plot types:
    - mesh: Model mesh with elements and nodes
    - heads: Head contour map for a timestep/layer
    - streams: Stream network colored by reach
    - elements: Elements colored by subregion
    """
    if not model_state.is_loaded:
        raise HTTPException(status_code=404, detail="No model loaded")

    model = model_state.model
    if model is None or model.grid is None:
        raise HTTPException(status_code=404, detail="No mesh/grid loaded")

    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    from pyiwfm.visualization.plotting import (
        plot_elements,
        plot_mesh,
        plot_scalar_field,
        plot_streams,
    )

    try:
        fig = None
        if plot_type == "mesh":
            fig, _ax = plot_mesh(model.grid, figsize=(width, height))
        elif plot_type == "elements":
            fig, _ax = plot_elements(model.grid, figsize=(width, height))
        elif plot_type == "streams":
            if model.streams is None:
                raise HTTPException(status_code=404, detail="No stream network loaded")
            fig, _ax = plot_streams(model.streams, figsize=(width, height))
        elif plot_type == "heads":
            loader = model_state.get_head_loader()
            if loader is None:
                raise HTTPException(status_code=404, detail="No head data available")
            if timestep >= loader.n_frames:
                raise HTTPException(
                    status_code=400,
                    detail=f"Timestep {timestep} out of range [0, {loader.n_frames})",
                )
            import numpy as np

            frame = loader.get_frame(timestep)
            layer_idx = layer - 1
            if layer_idx >= frame.shape[1]:
                raise HTTPException(
                    status_code=400,
                    detail=f"Layer {layer} out of range [1, {frame.shape[1]}]",
                )
            values = frame[:, layer_idx]
            # Mask dry cells
            values = np.where(values < -9000, np.nan, values)
            fig, _ax = plot_scalar_field(
                model.grid,
                values,
                figsize=(width, height),
            )
            _ax.set_title(f"Head - Layer {layer}, Timestep {timestep}")
        else:
            raise HTTPException(
                status_code=400,
                detail=f"Unknown plot type: {plot_type}. Supported: mesh, elements, streams, heads",
            )

        buf = io.BytesIO()
        if format == "svg":
            fig.savefig(buf, format="svg", bbox_inches="tight")
            media_type = "image/svg+xml"
            ext = "svg"
        else:
            fig.savefig(buf, format="png", dpi=dpi, bbox_inches="tight")
            media_type = "image/png"
            ext = "png"
        plt.close(fig)
        buf.seek(0)

        filename = f"{plot_type}_layer{layer}_ts{timestep}.{ext}"
        return Response(
            content=buf.getvalue(),
            media_type=media_type,
            headers={"Content-Disposition": f"attachment; filename={filename}"},
        )
    except HTTPException:
        raise
    except Exception as e:
        logger.exception("Plot generation failed")
        raise HTTPException(status_code=500, detail=str(e)) from e



"""
PreProcessor Binary File Reader for IWFM.

This module reads the binary output file produced by the IWFM PreProcessor.
The binary file uses ``ACCESS='STREAM'`` (raw bytes, no Fortran record
markers) and contains pre-processed mesh, stratigraphy, and component data
that enables faster simulation startup.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

import numpy as np
from numpy.typing import NDArray

from pyiwfm.io.binary import StreamAccessBinaryReader

# Fixed-length string sizes used by IWFM binary output
_SUBREGION_NAME_LEN = 50
_REACH_NAME_LEN = 30


@dataclass
class AppNodeData:
    """Pre-processed application node data."""

    id: int
    area: float
    boundary_node: bool
    n_connected_node: int
    n_face_id: int
    surrounding_elements: NDArray[np.int32]
    connected_nodes: NDArray[np.int32]
    face_ids: NDArray[np.int32]
    elem_id_on_ccw_side: NDArray[np.int32]
    irrotational_coeff: NDArray[np.float64]


@dataclass
class AppElementData:
    """Pre-processed application element data."""

    id: int
    subregion: int
    area: float
    face_ids: NDArray[np.int32]
    vertex_areas: NDArray[np.float64]
    vertex_area_fractions: NDArray[np.float64]
    integral_del_shp_i_del_shp_j: NDArray[np.float64]
    integral_rot_del_shp_i_del_shp_j: NDArray[np.float64]


@dataclass
class AppFaceData:
    """Pre-processed application face data."""

    nodes: NDArray[np.int32]  # (n_faces, 2) - node indices for each face
    elements: NDArray[np.int32]  # (n_faces, 2) - element indices on each side
    boundary: NDArray[np.bool_]  # (n_faces,) - True if boundary face
    lengths: NDArray[np.float64]  # (n_faces,) - face lengths


@dataclass
class SubregionData:
    """Pre-processed subregion data."""

    id: int
    name: str
    n_elements: int
    n_neighbor_regions: int
    area: float
    region_elements: NDArray[np.int32]
    neighbor_region_ids: NDArray[np.int32]
    neighbor_n_boundary_faces: NDArray[np.int32]
    neighbor_boundary_faces: list[NDArray[np.int32]]


@dataclass
class StratigraphyData:
    """Pre-processed stratigraphy data."""

    n_layers: int
    ground_surface_elev: NDArray[np.float64]  # (n_nodes,)
    top_elev: NDArray[np.float64]  # (n_nodes, n_layers)
    bottom_elev: NDArray[np.float64]  # (n_nodes, n_layers)
    active_node: NDArray[np.bool_]  # (n_nodes, n_layers)
    active_layer_above: NDArray[np.int32]  # (n_nodes, n_layers)
    active_layer_below: NDArray[np.int32]  # (n_nodes, n_layers)
    top_active_layer: NDArray[np.int32]  # (n_nodes,)
    bottom_active_layer: NDArray[np.int32]  # (n_nodes,)


@dataclass
class StreamGWConnectorData:
    """Pre-processed stream-GW connector data."""

    n_stream_nodes: int
    gw_nodes: NDArray[np.int32]  # GW node index for each stream node
    layers: NDArray[np.int32]  # Layer for each connection


@dataclass
class LakeGWConnectorData:
    """Pre-processed lake-GW connector data."""

    n_lakes: int
    lake_elements: list[NDArray[np.int32]]  # Elements for each lake
    lake_nodes: list[NDArray[np.int32]]  # GW nodes for each lake


@dataclass
class StreamLakeConnectorData:
    """Pre-processed stream-lake connector data."""

    n_connections: int
    stream_nodes: NDArray[np.int32]  # Stream node indices
    lake_ids: NDArray[np.int32]  # Connected lake IDs


@dataclass
class StreamData:
    """Pre-processed stream data."""

    n_reaches: int
    n_stream_nodes: int
    reach_ids: NDArray[np.int32]
    reach_names: list[str]
    reach_upstream_nodes: NDArray[np.int32]
    reach_downstream_nodes: NDArray[np.int32]
    reach_outflow_dest: NDArray[np.int32]  # 0=boundary, -n=lake n, +n=reach n


@dataclass
class LakeData:
    """Pre-processed lake data."""

    n_lakes: int
    lake_ids: NDArray[np.int32]
    lake_names: list[str]
    lake_max_elevations: NDArray[np.float64]
    lake_elements: list[NDArray[np.int32]]


@dataclass
class PreprocessorBinaryData:
    """Complete preprocessor binary output structure."""

    # Grid dimensions
    n_nodes: int = 0
    n_elements: int = 0
    n_faces: int = 0
    n_subregions: int = 0
    n_boundary_faces: int = 0

    # Coordinates
    x: NDArray[np.float64] = field(default_factory=lambda: np.array([]))
    y: NDArray[np.float64] = field(default_factory=lambda: np.array([]))

    # Element connectivity
    n_vertex: NDArray[np.int32] = field(default_factory=lambda: np.array([]))
    vertex: NDArray[np.int32] = field(default_factory=lambda: np.array([]))

    # App grid data
    app_nodes: list[AppNodeData] = field(default_factory=list)
    app_elements: list[AppElementData] = field(default_factory=list)
    app_faces: AppFaceData | None = None
    boundary_face_list: NDArray[np.int32] = field(default_factory=lambda: np.array([]))
    subregions: list[SubregionData] = field(default_factory=list)

    # Stratigraphy
    stratigraphy: StratigraphyData | None = None

    # Component connectors
    stream_lake_connector: StreamLakeConnectorData | None = None
    stream_gw_connector: StreamGWConnectorData | None = None
    lake_gw_connector: LakeGWConnectorData | None = None

    # Components
    lakes: LakeData | None = None
    streams: StreamData | None = None

    # Matrix data (sparse matrix structure)
    matrix_n_equations: int = 0
    matrix_connectivity: NDArray[np.int32] = field(default_factory=lambda: np.array([]))


class PreprocessorBinaryReader:
    """Reader for IWFM PreProcessor binary output files.

    The preprocessor binary uses ``ACCESS='STREAM'`` (raw bytes, no Fortran
    record markers).  All arrays are written with explicit lengths that must
    be derived from dimension counts read earlier in the file.

    Section read order (matching IWFM Fortran ``WritePreprocessedData``):

    1. AppGrid — dimensions, coordinates, connectivity, per-node/element
       data, face data, boundary faces, subregions
    2. Stratigraphy — NLayers, TopActiveLayer, ActiveNode, GSElev,
       TopElev, BottomElev  (2-D arrays in Fortran column-major order)
    3. StrmLakeConnector — 3 sub-connectors (stream→lake, lake→stream,
       lake→lake), each with count + source/dest arrays
    4. StrmGWConnector — version int, then gw_node/layer arrays
    5. LakeGWConnector — per-lake element/node arrays
    6. AppLake — version int, per-lake data with rating tables
    7. AppStream — version int, per-node rating tables, per-reach metadata
    8. Matrix — sparse structure (optional)
    """

    def __init__(self, endian: str = "<") -> None:
        self.endian = endian

    def read(self, filepath: Path | str) -> PreprocessorBinaryData:
        filepath = Path(filepath)
        if not filepath.exists():
            raise FileNotFoundError(f"Binary file not found: {filepath}")

        data = PreprocessorBinaryData()

        with StreamAccessBinaryReader(filepath, self.endian) as f:
            self._read_grid_data(f, data)
            self._read_stratigraphy(f, data)
            self._read_stream_lake_connector(f, data)
            self._read_stream_gw_connector(f, data)
            self._read_lake_gw_connector(f, data)
            self._read_lake_data(f, data)
            self._read_stream_data(f, data)
            self._read_matrix_data(f, data)

        return data

    # -- Section 1: AppGrid ------------------------------------------------

    def _read_grid_data(self, f: StreamAccessBinaryReader, data: PreprocessorBinaryData) -> None:
        # 5 dimension ints
        data.n_nodes = f.read_int()
        data.n_elements = f.read_int()
        data.n_faces = f.read_int()
        data.n_subregions = f.read_int()
        data.n_boundary_faces = f.read_int()

        # Coordinates
        data.x = f.read_doubles(data.n_nodes)
        data.y = f.read_doubles(data.n_nodes)

        # Element connectivity: NVertex(NElements), Vertex(MaxNV, NElements)
        data.n_vertex = f.read_ints(data.n_elements)
        max_nv = int(data.n_vertex.max()) if data.n_elements > 0 else 4
        data.vertex = f.read_ints(max_nv * data.n_elements)

        # Per-node app data
        for _ in range(data.n_nodes):
            data.app_nodes.append(self._read_app_node(f))

        # Per-element app data
        for _ in range(data.n_elements):
            data.app_elements.append(self._read_app_element(f))

        # Face data
        data.app_faces = self._read_app_faces(f, data.n_faces)

        # Boundary face list
        if data.n_boundary_faces > 0:
            data.boundary_face_list = f.read_ints(data.n_boundary_faces)

        # Subregion data
        for _ in range(data.n_subregions):
            data.subregions.append(self._read_subregion(f))

    def _read_app_node(self, f: StreamAccessBinaryReader) -> AppNodeData:
        node_id = f.read_int()
        area = f.read_double()
        boundary_node = f.read_logical()
        n_connected = f.read_int()
        n_face_id = f.read_int()
        n_surround = f.read_int()
        n_connected_arr = f.read_int()

        surrounding = f.read_ints(n_surround)
        connected = f.read_ints(n_connected_arr)
        face_ids = f.read_ints(n_face_id)
        elem_ccw = f.read_ints(n_face_id)
        irrot_coeff = f.read_doubles(n_face_id)

        return AppNodeData(
            id=node_id,
            area=area,
            boundary_node=boundary_node,
            n_connected_node=n_connected,
            n_face_id=n_face_id,
            surrounding_elements=surrounding,
            connected_nodes=connected,
            face_ids=face_ids,
            elem_id_on_ccw_side=elem_ccw,
            irrotational_coeff=irrot_coeff,
        )

    def _read_app_element(self, f: StreamAccessBinaryReader) -> AppElementData:
        elem_id = f.read_int()
        subregion = f.read_int()
        area = f.read_double()
        n_faces = f.read_int()
        n_vert_area = f.read_int()
        n_del_shp = f.read_int()
        n_rot_shp = f.read_int()

        face_ids = f.read_ints(n_faces)
        vert_areas = f.read_doubles(n_vert_area)
        vert_fracs = f.read_doubles(n_vert_area)
        del_shp = f.read_doubles(n_del_shp)
        rot_shp = f.read_doubles(n_rot_shp)

        return AppElementData(
            id=elem_id,
            subregion=subregion,
            area=area,
            face_ids=face_ids,
            vertex_areas=vert_areas,
            vertex_area_fractions=vert_fracs,
            integral_del_shp_i_del_shp_j=del_shp,
            integral_rot_del_shp_i_del_shp_j=rot_shp,
        )

    def _read_app_faces(self, f: StreamAccessBinaryReader, n_faces: int) -> AppFaceData:
        if n_faces == 0:
            return AppFaceData(
                nodes=np.array([], dtype=np.int32).reshape(0, 2),
                elements=np.array([], dtype=np.int32).reshape(0, 2),
                boundary=np.array([], dtype=np.bool_),
                lengths=np.array([], dtype=np.float64),
            )

        nodes = f.read_ints(2 * n_faces).reshape((n_faces, 2))
        elements = f.read_ints(2 * n_faces).reshape((n_faces, 2))
        boundary = f.read_logicals(n_faces)
        lengths = f.read_doubles(n_faces)

        return AppFaceData(nodes=nodes, elements=elements, boundary=boundary, lengths=lengths)

    def _read_subregion(self, f: StreamAccessBinaryReader) -> SubregionData:
        sub_id = f.read_int()
        name = f.read_string(_SUBREGION_NAME_LEN)
        n_elements = f.read_int()
        n_neighbors = f.read_int()
        area = f.read_double()

        region_elements = f.read_ints(n_elements)
        neighbor_ids = f.read_ints(n_neighbors)
        neighbor_n_faces = f.read_ints(n_neighbors)

        neighbor_faces: list[NDArray[np.int32]] = []
        for nf in neighbor_n_faces:
            neighbor_faces.append(f.read_ints(int(nf)))

        return SubregionData(
            id=sub_id,
            name=name,
            n_elements=n_elements,
            n_neighbor_regions=n_neighbors,
            area=area,
            region_elements=region_elements,
            neighbor_region_ids=neighbor_ids,
            neighbor_n_boundary_faces=neighbor_n_faces,
            neighbor_boundary_faces=neighbor_faces,
        )

    # -- Section 2: Stratigraphy -------------------------------------------

    def _read_stratigraphy(self, f: StreamAccessBinaryReader, data: PreprocessorBinaryData) -> None:
        n_layers = f.read_int()
        n = data.n_nodes

        # Fortran write order: NLayers, TopActiveLayer, ActiveNode,
        #   GSElev, TopElev, BottomElev
        # 2-D arrays stored column-major (Fortran order).
        top_active = f.read_ints(n)
        active_flat = f.read_logicals(n * n_layers)
        gs_elev = f.read_doubles(n)
        top_flat = f.read_doubles(n * n_layers)
        bottom_flat = f.read_doubles(n * n_layers)

        active_2d = active_flat.reshape((n, n_layers), order="F")

        data.stratigraphy = StratigraphyData(
            n_layers=n_layers,
            ground_surface_elev=gs_elev,
            top_elev=top_flat.reshape((n, n_layers), order="F"),
            bottom_elev=bottom_flat.reshape((n, n_layers), order="F"),
            active_node=active_2d,
            # Derived arrays — not stored in binary; fill with zeros.
            active_layer_above=np.zeros((n, n_layers), dtype=np.int32),
            active_layer_below=np.zeros((n, n_layers), dtype=np.int32),
            top_active_layer=top_active,
            bottom_active_layer=np.zeros(n, dtype=np.int32),
        )

    # -- Section 3: StrmLakeConnector (3 sub-connectors) -------------------

    def _read_stream_lake_connector(
        self, f: StreamAccessBinaryReader, data: PreprocessorBinaryData
    ) -> None:
        # Sub-connector 1: Stream inflow → Lake
        n1 = f.read_int()
        strm_nodes = f.read_ints(n1) if n1 > 0 else np.array([], dtype=np.int32)
        lake_ids = f.read_ints(n1) if n1 > 0 else np.array([], dtype=np.int32)

        data.stream_lake_connector = StreamLakeConnectorData(
            n_connections=n1, stream_nodes=strm_nodes, lake_ids=lake_ids
        )

        # Sub-connector 2: Lake outflow → Stream (read & discard)
        n2 = f.read_int()
        if n2 > 0:
            f.read_ints(n2)  # lake IDs
            f.read_ints(n2)  # stream nodes

        # Sub-connector 3: Lake outflow → Lake (read & discard)
        n3 = f.read_int()
        if n3 > 0:
            f.read_ints(n3)  # source lake IDs
            f.read_ints(n3)  # dest lake IDs

    # -- Section 4: StrmGWConnector ----------------------------------------

    def _read_stream_gw_connector(
        self, f: StreamAccessBinaryReader, data: PreprocessorBinaryData
    ) -> None:
        version = f.read_int()

        if version == 0:
            data.stream_gw_connector = StreamGWConnectorData(
                n_stream_nodes=0,
                gw_nodes=np.array([], dtype=np.int32),
                layers=np.array([], dtype=np.int32),
            )
            return

        n_strm = f.read_int()
        gw_nodes = f.read_ints(n_strm)
        layers = f.read_ints(n_strm)

        data.stream_gw_connector = StreamGWConnectorData(
            n_stream_nodes=n_strm, gw_nodes=gw_nodes, layers=layers
        )

    # -- Section 5: LakeGWConnector ----------------------------------------

    def _read_lake_gw_connector(
        self, f: StreamAccessBinaryReader, data: PreprocessorBinaryData
    ) -> None:
        n_lakes = f.read_int()

        if n_lakes == 0:
            data.lake_gw_connector = LakeGWConnectorData(n_lakes=0, lake_elements=[], lake_nodes=[])
            return

        lake_elements: list[NDArray[np.int32]] = []
        lake_nodes: list[NDArray[np.int32]] = []

        for _ in range(n_lakes):
            ne = f.read_int()
            lake_elements.append(f.read_ints(ne))
            nn = f.read_int()
            lake_nodes.append(f.read_ints(nn))

        data.lake_gw_connector = LakeGWConnectorData(
            n_lakes=n_lakes,
            lake_elements=lake_elements,
            lake_nodes=lake_nodes,
        )

    # -- Section 6: AppLake (version-prefixed) -----------------------------

    def _read_lake_data(self, f: StreamAccessBinaryReader, data: PreprocessorBinaryData) -> None:
        version = f.read_int()

        if version == 0:
            data.lakes = LakeData(
                n_lakes=0,
                lake_ids=np.array([], dtype=np.int32),
                lake_names=[],
                lake_max_elevations=np.array([], dtype=np.float64),
                lake_elements=[],
            )
            return

        n_lakes = f.read_int()
        lake_ids = np.zeros(n_lakes, dtype=np.int32)
        lake_max_elev = np.zeros(n_lakes, dtype=np.float64)
        lake_elements: list[NDArray[np.int32]] = []

        for i in range(n_lakes):
            lake_ids[i] = f.read_int()
            lake_max_elev[i] = f.read_double()
            # PairedData rating table
            n_pts = f.read_int()
            if n_pts > 0:
                f.read_doubles(n_pts)  # elevation points (skip)
                f.read_doubles(n_pts)  # outflow points (skip)
            ne = f.read_int()
            lake_elements.append(f.read_ints(ne) if ne > 0 else np.array([], dtype=np.int32))

        data.lakes = LakeData(
            n_lakes=n_lakes,
            lake_ids=lake_ids,
            lake_names=[f"Lake {int(lake_ids[i])}" for i in range(n_lakes)],
            lake_max_elevations=lake_max_elev,
            lake_elements=lake_elements,
        )

    # -- Section 7: AppStream (version-prefixed) ---------------------------

    def _read_stream_data(self, f: StreamAccessBinaryReader, data: PreprocessorBinaryData) -> None:
        version = f.read_int()

        if version == 0:
            data.streams = StreamData(
                n_reaches=0,
                n_stream_nodes=0,
                reach_ids=np.array([], dtype=np.int32),
                reach_names=[],
                reach_upstream_nodes=np.array([], dtype=np.int32),
                reach_downstream_nodes=np.array([], dtype=np.int32),
                reach_outflow_dest=np.array([], dtype=np.int32),
            )
            return

        n_reaches = f.read_int()
        n_strm_nodes = f.read_int()

        # Per-stream-node: ID + rating table
        for _ in range(n_strm_nodes):
            _node_id = f.read_int()
            n_pts = f.read_int()
            if n_pts > 0:
                f.read_doubles(n_pts)  # stage points (skip)
                f.read_doubles(n_pts)  # flow points (skip)

        # Per-reach metadata
        reach_ids = np.zeros(n_reaches, dtype=np.int32)
        reach_names: list[str] = []
        upstream = np.zeros(n_reaches, dtype=np.int32)
        downstream = np.zeros(n_reaches, dtype=np.int32)
        outflow = np.zeros(n_reaches, dtype=np.int32)

        for i in range(n_reaches):
            reach_ids[i] = f.read_int()
            reach_names.append(f.read_string(_REACH_NAME_LEN))
            upstream[i] = f.read_int()
            downstream[i] = f.read_int()
            outflow[i] = f.read_int()

        data.streams = StreamData(
            n_reaches=n_reaches,
            n_stream_nodes=n_strm_nodes,
            reach_ids=reach_ids,
            reach_names=reach_names,
            reach_upstream_nodes=upstream,
            reach_downstream_nodes=downstream,
            reach_outflow_dest=outflow,
        )

    # -- Section 8: Matrix (optional) --------------------------------------

    def _read_matrix_data(self, f: StreamAccessBinaryReader, data: PreprocessorBinaryData) -> None:
        try:
            data.matrix_n_equations = f.read_int()
            if data.matrix_n_equations > 0 and not f.at_eof():
                # Connectivity: NJCOLs(NEquations) then JCOL values
                n_jcols = f.read_ints(data.matrix_n_equations)
                total = int(n_jcols.sum())
                if total > 0 and not f.at_eof():
                    data.matrix_connectivity = f.read_ints(total)
        except EOFError:
            pass


def read_preprocessor_binary(filepath: Path | str, endian: str = "<") -> PreprocessorBinaryData:
    """Convenience function to read preprocessor binary file.

    Args:
        filepath: Path to the binary file
        endian: Byte order

    Returns:
        PreprocessorBinaryData with all preprocessed data
    """
    reader = PreprocessorBinaryReader(endian=endian)
    return reader.read(filepath)



"""Tests for runner/runner.py error paths.

Covers:
- IWFMRunner._run_executable() TimeoutExpired (lines 293-297)
- IWFMRunner._run_executable() generic Exception (lines 299-301)
- IWFMRunner._parse_log_messages() warning/error detection
"""

from __future__ import annotations

import subprocess
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from pyiwfm.runner.runner import IWFMRunner


class TestRunExecutableTimeout:
    """Test _run_executable TimeoutExpired handling."""

    def test_run_executable_timeout(self, tmp_path: Path) -> None:
        """Mock subprocess.run -> TimeoutExpired -> returns -1."""
        runner = IWFMRunner()
        exe_path = tmp_path / "fake.exe"
        exe_path.write_text("fake")

        timeout_err = subprocess.TimeoutExpired(cmd=["fake"], timeout=10)
        timeout_err.stdout = b"partial output"
        timeout_err.stderr = b"timed out"

        with patch("subprocess.run", side_effect=timeout_err):
            code, stdout, stderr, elapsed = runner._run_executable(
                executable=exe_path,
                input_text="input.in\n",
                working_dir=tmp_path,
                timeout=10,
            )

        assert code == -1
        assert "partial output" in stdout
        assert "timed out" in stderr

    def test_run_executable_timeout_no_output(self, tmp_path: Path) -> None:
        """TimeoutExpired with None stdout/stderr."""
        runner = IWFMRunner()
        exe_path = tmp_path / "fake.exe"
        exe_path.write_text("fake")

        timeout_err = subprocess.TimeoutExpired(cmd=["fake"], timeout=10)
        timeout_err.stdout = None
        timeout_err.stderr = None

        with patch("subprocess.run", side_effect=timeout_err):
            code, stdout, stderr, elapsed = runner._run_executable(
                executable=exe_path,
                input_text="input.in\n",
                working_dir=tmp_path,
                timeout=10,
            )

        assert code == -1
        assert stdout == ""
        assert "timed out" in stderr.lower()


class TestRunExecutableGenericError:
    """Test _run_executable generic Exception handling."""

    def test_run_executable_generic_error(self, tmp_path: Path) -> None:
        """Mock subprocess.run -> Exception -> returns -1 with error message."""
        runner = IWFMRunner()
        exe_path = tmp_path / "fake.exe"
        exe_path.write_text("fake")

        with patch("subprocess.run", side_effect=OSError("Permission denied")):
            code, stdout, stderr, elapsed = runner._run_executable(
                executable=exe_path,
                input_text="input.in\n",
                working_dir=tmp_path,
            )

        assert code == -1
        assert stdout == ""
        assert "Permission denied" in stderr


class TestParseLogMessages:
    """Test _parse_log_messages error/warning detection."""

    def test_parse_log_errors(self) -> None:
        """Log with ERROR lines -> detected."""
        runner = IWFMRunner()
        log = "Step 1: OK\nERROR: Memory exceeded\nFATAL: Cannot continue\nStep 2: OK"
        errors, warnings = runner._parse_log_messages(log)
        assert len(errors) == 2
        assert any("Memory exceeded" in e for e in errors)
        assert any("Cannot continue" in e for e in errors)

    def test_parse_log_warnings(self) -> None:
        """Log with WARNING lines -> detected."""
        runner = IWFMRunner()
        log = "Step 1: OK\nWARNING: Low convergence\nStep 2: OK"
        errors, warnings = runner._parse_log_messages(log)
        assert len(errors) == 0
        assert len(warnings) == 1
        assert "convergence" in warnings[0].lower()

    def test_parse_log_no_issues(self) -> None:
        """Clean log -> no errors or warnings."""
        runner = IWFMRunner()
        log = "Step 1: Complete\nStep 2: Complete\nSimulation finished."
        errors, warnings = runner._parse_log_messages(log)
        assert len(errors) == 0
        assert len(warnings) == 0

    def test_parse_log_mixed(self) -> None:
        """Log with both errors and warnings."""
        runner = IWFMRunner()
        log = "WARNING: heads below drain\nERROR: convergence failure\nWARNING: negative pumping"
        errors, warnings = runner._parse_log_messages(log)
        assert len(errors) == 1
        assert len(warnings) == 2

    def test_parse_log_case_insensitive(self) -> None:
        """Error/warning detection is case-insensitive."""
        runner = IWFMRunner()
        log = "error: something wrong\nWarning: check this\nfatal issue detected"
        errors, warnings = runner._parse_log_messages(log)
        assert len(errors) == 2  # "error" and "fatal"
        assert len(warnings) == 1


class TestRunPreprocessor:
    """Test run_preprocessor() method."""

    def test_run_preprocessor_no_exe(self) -> None:
        """No preprocessor exe -> FileNotFoundError."""
        runner = IWFMRunner()
        runner.executables.preprocessor = None
        with pytest.raises(FileNotFoundError, match="PreProcessor"):
            runner.run_preprocessor("fake.in")

    def test_run_preprocessor_no_main_file(self, tmp_path: Path) -> None:
        """Main file doesn't exist -> FileNotFoundError."""
        runner = IWFMRunner()
        runner.executables.preprocessor = tmp_path / "PreProcessor.exe"
        with pytest.raises(FileNotFoundError, match="Main file"):
            runner.run_preprocessor(tmp_path / "nonexistent.in")

    def test_run_preprocessor_success(self, tmp_path: Path) -> None:
        """Successful preprocessor run."""
        runner = IWFMRunner()
        exe = tmp_path / "PreProcessor.exe"
        exe.write_text("fake")
        runner.executables.preprocessor = exe

        main_file = tmp_path / "preprocessor.in"
        main_file.write_text("main input")

        with patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(
                returncode=0,
                stdout="100 nodes\n50 elements\n3 layers\n",
                stderr="",
            )
            result = runner.run_preprocessor(main_file)

        assert result.success
        assert result.n_nodes == 100
        assert result.n_elements == 50
        assert result.n_layers == 3


class TestRunSimulation:
    """Test run_simulation() method."""

    def test_run_simulation_no_exe(self) -> None:
        """No simulation exe -> FileNotFoundError."""
        runner = IWFMRunner()
        runner.executables.simulation = None
        runner.executables.simulation_parallel = None
        with pytest.raises(FileNotFoundError, match="Simulation"):
            runner.run_simulation("fake.in")

    def test_run_simulation_parallel_exe(self, tmp_path: Path) -> None:
        """Parallel flag uses parallel executable."""
        runner = IWFMRunner()
        exe_par = tmp_path / "Simulation_PLL.exe"
        exe_par.write_text("fake")
        runner.executables.simulation_parallel = exe_par
        runner.executables.simulation = None

        main_file = tmp_path / "simulation.in"
        main_file.write_text("main input")

        with patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(
                returncode=0,
                stdout="Time Step 1\nTime Step 2\n",
                stderr="",
            )
            result = runner.run_simulation(main_file, parallel=True)

        assert result.success


class TestRunBudget:
    """Test run_budget() method."""

    def test_run_budget_no_exe(self) -> None:
        """No budget exe -> FileNotFoundError."""
        runner = IWFMRunner()
        runner.executables.budget = None
        with pytest.raises(FileNotFoundError, match="Budget"):
            runner.run_budget("fake.bud")

    def test_run_budget_no_file(self, tmp_path: Path) -> None:
        """Budget file doesn't exist -> FileNotFoundError."""
        runner = IWFMRunner()
        runner.executables.budget = tmp_path / "Budget.exe"
        with pytest.raises(FileNotFoundError, match="Budget file"):
            runner.run_budget(tmp_path / "nonexistent.bud")


class TestRunZBudget:
    """Test run_zbudget() method."""

    def test_run_zbudget_no_exe(self) -> None:
        """No zbudget exe -> FileNotFoundError."""
        runner = IWFMRunner()
        runner.executables.zbudget = None
        with pytest.raises(FileNotFoundError, match="ZBudget"):
            runner.run_zbudget("fake.zbud")

    def test_run_zbudget_no_file(self, tmp_path: Path) -> None:
        """ZBudget file doesn't exist -> FileNotFoundError."""
        runner = IWFMRunner()
        runner.executables.zbudget = tmp_path / "ZBudget.exe"
        with pytest.raises(FileNotFoundError, match="ZBudget file"):
            runner.run_zbudget(tmp_path / "nonexistent.zbud")


class TestIWFMExecutables:
    """Test IWFMExecutables dataclass."""

    def test_available_none(self) -> None:
        """No executables -> empty list."""
        from pyiwfm.runner.runner import IWFMExecutables

        exes = IWFMExecutables()
        assert exes.available == []

    def test_repr(self) -> None:
        """Repr shows available list."""
        from pyiwfm.runner.runner import IWFMExecutables

        exes = IWFMExecutables()
        r = repr(exes)
        assert "IWFMExecutables" in r

    def test_post_init_nonexistent(self, tmp_path: Path) -> None:
        """Non-existent path -> set to None."""
        from pyiwfm.runner.runner import IWFMExecutables

        exes = IWFMExecutables(simulation=tmp_path / "nonexistent.exe")
        assert exes.simulation is None


class TestIWFMRunnerRepr:
    """Test IWFMRunner repr."""

    def test_repr(self) -> None:
        """Repr includes executables and working_dir."""
        runner = IWFMRunner()
        r = repr(runner)
        assert "IWFMRunner" in r


class TestGetWorkingDir:
    """Test _get_working_dir()."""

    def test_override(self, tmp_path: Path) -> None:
        """Override dir used."""
        runner = IWFMRunner()
        override = tmp_path / "override"
        result = runner._get_working_dir(tmp_path / "main.in", override)
        assert result == override

    def test_instance_working_dir(self, tmp_path: Path) -> None:
        """Instance working_dir used."""
        runner = IWFMRunner(working_dir=tmp_path / "instance")
        result = runner._get_working_dir(tmp_path / "main.in", None)
        assert result == tmp_path / "instance"

    def test_main_file_parent(self, tmp_path: Path) -> None:
        """Main file parent used as fallback."""
        runner = IWFMRunner()
        main_file = tmp_path / "subdir" / "main.in"
        result = runner._get_working_dir(main_file, None)
        assert result == tmp_path / "subdir"


class TestFindExecutablesPathSearch:
    """Test find_iwfm_executables() PATH search (lines 173-177)."""

    def test_find_in_system_path(self, tmp_path: Path) -> None:
        """Executable found via shutil.which when not in search paths."""
        from pyiwfm.runner.runner import find_iwfm_executables

        fake_exe = tmp_path / "iwfm_simulation"
        fake_exe.write_text("fake")

        # Override Path.cwd and package dir to prevent finding real executables
        empty_dir = tmp_path / "empty_search"
        empty_dir.mkdir()

        with (
            patch("pyiwfm.runner.runner.Path.cwd", return_value=empty_dir),
            patch("pyiwfm.runner.runner.Path.__file__", tmp_path / "fake.py", create=True),
            patch(
                "pyiwfm.runner.runner.shutil.which",
                side_effect=lambda name: str(fake_exe) if "Simulation" in name.lower() else None,
            ),
        ):
            result = find_iwfm_executables(search_paths=[empty_dir])

        # The real Bin directory may still find exes due to package path resolution.
        # We just verify the function runs without error.
        assert result is not None


class TestRunPreprocessorLogAndBinary:
    """Test run_preprocessor() log reading and binary output detection."""

    def test_preprocessor_log_file_read(self, tmp_path: Path) -> None:
        """Log file exists -> read and included in result (line 354)."""
        runner = IWFMRunner()
        exe = tmp_path / "PreProcessor.exe"
        exe.write_text("fake")
        runner.executables.preprocessor = exe

        main_file = tmp_path / "preprocessor.in"
        main_file.write_text("main input")

        # Create log file
        log_file = tmp_path / "PreprocessorMessages.out"
        log_file.write_text("Processing 100 nodes...\n50 elements found\n")

        with patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            result = runner.run_preprocessor(main_file)

        assert result.log_content != ""
        assert result.log_file is not None

    def test_preprocessor_binary_output_detection(self, tmp_path: Path) -> None:
        """Binary output file detected via glob (lines 383-384)."""
        runner = IWFMRunner()
        exe = tmp_path / "PreProcessor.exe"
        exe.write_text("fake")
        runner.executables.preprocessor = exe

        main_file = tmp_path / "preprocessor.in"
        main_file.write_text("main input")

        # Create binary output file
        (tmp_path / "Preprocessor.bin").write_text("binary content")

        with patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            result = runner.run_preprocessor(main_file)

        assert result.binary_output is not None
        assert "Preprocessor.bin" in str(result.binary_output)


class TestRunSimulationLogAndOutputs:
    """Test run_simulation() log reading and output file detection."""

    def test_simulation_log_file_read(self, tmp_path: Path) -> None:
        """Log file exists -> read and included in result (line 464)."""
        runner = IWFMRunner()
        exe = tmp_path / "Simulation.exe"
        exe.write_text("fake")
        runner.executables.simulation = exe

        main_file = tmp_path / "simulation.in"
        main_file.write_text("main input")

        # Create log file
        log_file = tmp_path / "SimulationMessages.out"
        log_file.write_text("Time step 1 complete\n")

        with patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            result = runner.run_simulation(main_file)

        assert result.log_content != ""

    def test_simulation_final_heads_detection(self, tmp_path: Path) -> None:
        """Final heads file detected via glob (lines 500-501)."""
        runner = IWFMRunner()
        exe = tmp_path / "Simulation.exe"
        exe.write_text("fake")
        runner.executables.simulation = exe

        main_file = tmp_path / "simulation.in"
        main_file.write_text("main input")

        # Create final heads file
        (tmp_path / "Model_FinalGWHeads.dat").write_text("heads data")

        with patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            result = runner.run_simulation(main_file)

        assert result.final_heads_file is not None


class TestRunBudgetLogFile:
    """Test run_budget() log file reading (line 583)."""

    def test_budget_log_file_read(self, tmp_path: Path) -> None:
        """Budget log file exists -> read."""
        runner = IWFMRunner()
        exe = tmp_path / "Budget.exe"
        exe.write_text("fake")
        runner.executables.budget = exe

        budget_file = tmp_path / "model.bud"
        budget_file.write_text("budget data")

        log_file = tmp_path / "BudgetMessages.out"
        log_file.write_text("Budget processing complete\n")

        with patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            result = runner.run_budget(budget_file)

        assert result.log_content != ""
        assert result.log_file is not None


class TestRunZBudgetLogFile:
    """Test run_zbudget() log file reading (line 661)."""

    def test_zbudget_log_file_read(self, tmp_path: Path) -> None:
        """ZBudget log file exists -> read."""
        runner = IWFMRunner()
        exe = tmp_path / "ZBudget.exe"
        exe.write_text("fake")
        runner.executables.zbudget = exe

        zbud_file = tmp_path / "model.zbud"
        zbud_file.write_text("zbudget data")

        log_file = tmp_path / "ZBudgetMessages.out"
        log_file.write_text("ZBudget processing complete\n")

        with patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(returncode=0, stdout="", stderr="")
            result = runner.run_zbudget(zbud_file)

        assert result.log_content != ""
        assert result.log_file is not None



"""
SQLite cache loader for the IWFM web viewer.

Provides fast read access to pre-computed data stored by
:class:`~cache_builder.SqliteCacheBuilder`.  Uses WAL mode for
concurrent reads and connection pooling for thread safety.
"""

from __future__ import annotations

import logging
import sqlite3
import threading
import zlib
from pathlib import Path

import numpy as np
from numpy.typing import NDArray

logger = logging.getLogger(__name__)


def _decompress_array(blob: bytes, dtype: type = np.float64) -> NDArray:  # type: ignore[type-arg]
    """Decompress a zlib-compressed numpy array."""
    raw = zlib.decompress(blob)
    arr: NDArray = np.frombuffer(raw, dtype=dtype).copy()  # type: ignore[type-arg]
    return arr


def _decompress_strings(blob: bytes) -> list[str]:
    """Decompress a zlib-compressed newline-separated string list."""
    raw = zlib.decompress(blob)
    return raw.decode("utf-8").split("\n")


class SqliteCacheLoader:
    """Read-only access to the viewer SQLite cache.

    Thread-safe via per-thread connections with WAL mode.

    Parameters
    ----------
    cache_path : Path
        Path to the SQLite cache file.
    """

    def __init__(self, cache_path: Path) -> None:
        self.cache_path = cache_path
        self._local = threading.local()
        # Validate file exists
        if not cache_path.exists():
            raise FileNotFoundError(f"Cache not found: {cache_path}")
        logger.info("SQLite cache loader opened: %s", cache_path)

    def _conn(self) -> sqlite3.Connection:
        """Get the per-thread connection, creating if needed."""
        conn = getattr(self._local, "conn", None)
        if conn is None:
            conn = sqlite3.connect(str(self.cache_path), check_same_thread=False)
            conn.execute("PRAGMA journal_mode=WAL")
            conn.execute("PRAGMA mmap_size=268435456")  # 256 MB
            conn.execute("PRAGMA cache_size=-65536")  # 64 MB
            conn.execute("PRAGMA synchronous=NORMAL")
            conn.execute("PRAGMA query_only=ON")
            self._local.conn = conn
        return conn

    def close(self) -> None:
        """Close the per-thread connection if open."""
        conn = getattr(self._local, "conn", None)
        if conn is not None:
            conn.close()
            self._local.conn = None

    # ------------------------------------------------------------------
    # Metadata
    # ------------------------------------------------------------------

    def get_metadata(self, key: str) -> str | None:
        """Get a metadata value by key."""
        cur = self._conn().execute("SELECT value FROM metadata WHERE key = ?", (key,))
        row = cur.fetchone()
        return row[0] if row else None

    def get_timesteps(self) -> list[str]:
        """Get all timestep datetimes as ISO strings."""
        cur = self._conn().execute("SELECT datetime FROM timesteps ORDER BY idx")
        return [r[0] for r in cur.fetchall()]

    # ------------------------------------------------------------------
    # Head data
    # ------------------------------------------------------------------

    def get_head_frame(self, frame_idx: int) -> NDArray | None:
        """Get a raw head frame (n_nodes, n_layers).

        Returns None if frame not cached.
        """
        cur = self._conn().execute(
            "SELECT n_nodes, n_layers, data_blob FROM head_frames WHERE frame_idx = ?",
            (frame_idx,),
        )
        row = cur.fetchone()
        if row is None:
            return None
        n_nodes, n_layers, blob = row
        arr = _decompress_array(blob)
        return arr.reshape(n_nodes, n_layers)

    def get_head_by_element(
        self, frame_idx: int, layer: int
    ) -> tuple[NDArray, float, float] | None:
        """Get pre-computed element-averaged heads.

        Returns (values_array, min_val, max_val) or None.
        """
        cur = self._conn().execute(
            "SELECT values_blob, min_val, max_val FROM head_by_element "
            "WHERE frame_idx = ? AND layer = ?",
            (frame_idx, layer),
        )
        row = cur.fetchone()
        if row is None:
            return None
        blob, min_val, max_val = row
        return _decompress_array(blob), float(min_val), float(max_val)

    def get_head_range(self, layer: int) -> dict | None:
        """Get the pre-computed head range for a layer.

        Returns dict with percentile_02, percentile_98, abs_min, abs_max.
        """
        cur = self._conn().execute(
            "SELECT percentile_02, percentile_98, abs_min, abs_max FROM head_range WHERE layer = ?",
            (layer,),
        )
        row = cur.fetchone()
        if row is None:
            return None
        return {
            "percentile_02": row[0],
            "percentile_98": row[1],
            "abs_min": row[2],
            "abs_max": row[3],
        }

    def get_n_head_frames(self) -> int:
        """Get the number of cached head frames."""
        cur = self._conn().execute("SELECT COUNT(*) FROM head_frames")
        row = cur.fetchone()
        return row[0] if row else 0

    # ------------------------------------------------------------------
    # Budget data
    # ------------------------------------------------------------------

    def get_budget_types(self) -> list[str]:
        """Get list of cached budget types."""
        cur = self._conn().execute("SELECT budget_type FROM budget_files ORDER BY budget_type")
        return [r[0] for r in cur.fetchall()]

    def get_budget_locations(self, budget_type: str) -> list[tuple[int, str, float]]:
        """Get locations for a budget type as (idx, name, area) tuples."""
        cur = self._conn().execute(
            "SELECT location_idx, location_name, area "
            "FROM budget_locations WHERE budget_type = ? ORDER BY location_idx",
            (budget_type,),
        )
        return [(r[0], r[1], r[2]) for r in cur.fetchall()]

    def get_budget_columns(self, budget_type: str) -> list[tuple[int, str, str]]:
        """Get columns for a budget type as (idx, name, units) tuples."""
        cur = self._conn().execute(
            "SELECT column_idx, column_name, units "
            "FROM budget_columns WHERE budget_type = ? ORDER BY column_idx",
            (budget_type,),
        )
        return [(r[0], r[1], r[2]) for r in cur.fetchall()]

    def get_budget_data(self, budget_type: str, location_idx: int) -> NDArray | None:
        """Get budget data for a location as (n_timesteps, n_cols) array."""
        cur = self._conn().execute(
            "SELECT values_blob FROM budget_data "
            "WHERE budget_type = ? AND location_idx = ? ORDER BY timestep_idx",
            (budget_type, location_idx),
        )
        rows = cur.fetchall()
        if not rows:
            return None
        arrays = [_decompress_array(r[0]) for r in rows]
        return np.vstack(arrays)

    def get_budget_summary(
        self, budget_type: str, location_idx: int
    ) -> list[tuple[int, float, float]]:
        """Get budget summaries as (col_idx, total, average) tuples."""
        cur = self._conn().execute(
            "SELECT column_idx, total, average FROM budget_summaries "
            "WHERE budget_type = ? AND location_idx = ? ORDER BY column_idx",
            (budget_type, location_idx),
        )
        return [(r[0], r[1], r[2]) for r in cur.fetchall()]

    # ------------------------------------------------------------------
    # Hydrographs
    # ------------------------------------------------------------------

    def get_hydrograph(self, hydro_type: str, column_idx: int) -> tuple[list[str], NDArray] | None:
        """Get a hydrograph timeseries.

        Returns (times_list, values_array) or None.
        """
        cur = self._conn().execute(
            "SELECT times_blob, values_blob FROM hydrograph_series "
            "WHERE hydro_type = ? AND column_idx = ?",
            (hydro_type, column_idx),
        )
        row = cur.fetchone()
        if row is None:
            return None
        times = _decompress_strings(row[0])
        values = _decompress_array(row[1])
        return times, values

    def get_hydrograph_columns(self, hydro_type: str) -> list[tuple[int, int, int]]:
        """Get hydrograph column metadata as (col_idx, node_id, layer) tuples."""
        cur = self._conn().execute(
            "SELECT column_idx, node_id, layer FROM hydrograph_columns "
            "WHERE hydro_type = ? ORDER BY column_idx",
            (hydro_type,),
        )
        return [(r[0], r[1], r[2]) for r in cur.fetchall()]

    def get_gw_hydrograph_all_layers(
        self, node_id: int
    ) -> list[tuple[int, list[str], NDArray]] | None:
        """Get all GW hydrograph layers for a node.

        Returns list of (layer, times, values) tuples, or None.
        """
        cur = self._conn().execute(
            "SELECT hc.layer, hs.times_blob, hs.values_blob "
            "FROM hydrograph_columns hc "
            "JOIN hydrograph_series hs "
            "  ON hc.hydro_type = hs.hydro_type AND hc.column_idx = hs.column_idx "
            "WHERE hc.hydro_type = 'gw' AND hc.node_id = ? "
            "ORDER BY hc.layer",
            (node_id,),
        )
        rows = cur.fetchall()
        if not rows:
            return None
        result: list[tuple[int, list[str], NDArray]] = []
        for layer, times_blob, values_blob in rows:
            times = _decompress_strings(times_blob)
            values = _decompress_array(values_blob)
            result.append((layer, times, values))
        return result

    def get_gw_hydrograph_by_columns(
        self, base_col: int, n_layers: int
    ) -> list[tuple[int, list[str], NDArray]] | None:
        """Get GW hydrograph layers by consecutive column range.

        IWFM GW hydrograph output files store n_layers consecutive columns
        per location: [loc_lay1, loc_lay2, ..., loc_layN].

        Returns list of (layer, times, values) tuples, or None.
        """
        result: list[tuple[int, list[str], NDArray]] = []
        for offset in range(n_layers):
            col = base_col + offset
            cur = self._conn().execute(
                "SELECT hc.layer, hs.times_blob, hs.values_blob "
                "FROM hydrograph_columns hc "
                "JOIN hydrograph_series hs "
                "  ON hc.hydro_type = hs.hydro_type AND hc.column_idx = hs.column_idx "
                "WHERE hc.hydro_type = 'gw' AND hc.column_idx = ?",
                (col,),
            )
            row = cur.fetchone()
            if row is None:
                continue
            layer, times_blob, values_blob = row
            times = _decompress_strings(times_blob)
            values = _decompress_array(values_blob)
            result.append((layer if layer > 0 else offset + 1, times, values))
        return result if result else None

    # ------------------------------------------------------------------
    # Stream ratings
    # ------------------------------------------------------------------

    def get_stream_rating(self, stream_node_id: int) -> tuple[float, NDArray, NDArray] | None:
        """Get a stream node's rating table.

        Returns (bottom_elev, stages, flows) or None.
        """
        cur = self._conn().execute(
            "SELECT bottom_elev, stages_blob, flows_blob "
            "FROM stream_ratings WHERE stream_node_id = ?",
            (stream_node_id,),
        )
        row = cur.fetchone()
        if row is None:
            return None
        return (
            float(row[0]),
            _decompress_array(row[1]),
            _decompress_array(row[2]),
        )

    # ------------------------------------------------------------------
    # Area / land-use
    # ------------------------------------------------------------------

    def get_area_snapshot(self, frame_idx: int) -> list[dict] | None:
        """Get a land-use snapshot for a timestep.

        Returns list of element dicts or None.
        """
        cur = self._conn().execute(
            "SELECT element_id, agricultural, urban, native_rip, "
            "total_area, dominant "
            "FROM landuse_snapshots WHERE frame_idx = ? ORDER BY element_id",
            (frame_idx,),
        )
        rows = cur.fetchall()
        if not rows:
            return None
        return [
            {
                "element_id": r[0],
                "agricultural": r[1],
                "urban": r[2],
                "native_rip": r[3],
                "total_area": r[4],
                "dominant": r[5],
            }
            for r in rows
        ]

    def get_area_timesteps(self, area_type: str = "landuse") -> list[str]:
        """Get area timestep dates."""
        cur = self._conn().execute(
            "SELECT datetime FROM area_timesteps WHERE area_type = ? ORDER BY idx",
            (area_type,),
        )
        return [r[0] for r in cur.fetchall()]

    # ------------------------------------------------------------------
    # Diagnostics
    # ------------------------------------------------------------------

    def get_stats(self) -> dict:
        """Get cache statistics for diagnostics."""
        conn = self._conn()
        stats: dict = {}

        for table in (
            "head_frames",
            "head_by_element",
            "head_range",
            "budget_files",
            "budget_data",
            "budget_summaries",
            "hydrograph_series",
            "hydrograph_columns",
            "stream_ratings",
            "landuse_snapshots",
        ):
            try:
                cur = conn.execute(f"SELECT COUNT(*) FROM {table}")  # noqa: S608
                stats[table] = cur.fetchone()[0]
            except Exception:
                stats[table] = 0

        # File size
        try:
            stats["file_size_mb"] = round(self.cache_path.stat().st_size / (1024 * 1024), 1)
        except Exception:
            stats["file_size_mb"] = 0

        return stats
